#  Copyright 2024 CoreWeave
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import os
from typing import Optional
from P4 import P4, P4Exception

from p4utilsforunreal.log import get_logger


logger = get_logger()


class PerforceConnection:
    """
    Wrapper around the P4 object of p4python package

    P4 connection can be created by passing port, user, password and charset to the constructor
    or setting appropriate P4PORT, P4USER and P4PASSWD environment variables.

    .. note::
       Current connection properties will be used by default
    """

    def __init__(self, port: str = None, user: str = None, password: str = None, charset="none", required_connection=False):
        p4 = P4()
        p4.charset = charset
        self.p4_connection = False

        p4_port = port or os.getenv('P4PORT')
        if p4_port:
            p4.port = p4_port

        p4_user = user or os.getenv('P4USER')
        if p4_user:
            p4.user = p4_user

        p4_password = password or os.getenv('P4PASSWD')

        if required_connection:
            self._required_connection(p4, p4_port, p4_user, p4_password)
        else:
            self._connection(p4, p4_password)
        
    def _required_connection(self, p4, p4_port, p4_user, p4_password):
        """
        Attempt to connect to Perforce. Raise any exceptions to block
        Perforce-enabled MRQ submissions.

        :param p4: P4 connection object
        :param p4_port: Perforce port
        :param p4_user: Perforce username
        :param p4_password: Perforce password
        """

        if not all([p4_port, p4_user, p4_password]):
            p4_fields = {"Perforce Port": p4_port, "Perforce Username": p4_user, "Perforce Password": p4_password}
            p4_missing_fields = ", ".join([k for k, v in p4_fields.items() if v is None])
            raise Exception(f'Could not connect to Perforce server. Missing field(s) under Conductor Settings: {p4_missing_fields}.')

        try:
            p4.password = p4_password
            p4.connect()

            p4.input = 'y'
            p4.run('trust', ['-y'])
            p4.run_login()
        except Exception as e:
            p4e = str(e).lower()
            if "not connected" in p4e or "connect to server failed" in p4e:
                raise Exception(f'Could not connect to Perforce server. Invalid port ({p4.port}).')
            elif "password invalid" in p4e:
                raise Exception(f'Could not connect to Perforce server. Invalid username ({p4.user}) or password ({p4.password}).')
            raise Exception(e)

        self.p4 = p4
        self.p4_connection = True

    def _connection(self, p4, p4_password):
        """
        Attempt to connect to Perforce. Quietly log if any issues
        occur to not interefere with standard MRQ submissions.

        :param p4: P4 connection object
        :param p4_password: Perforce password
        """

        try:
            p4.connect()
        except P4Exception as e:
            logger.info(f'Could not connect Perforce server {p4.port} as user {p4.user}\n{str(e)}')

        p4.input = 'y'
        p4.run('trust', ['-y'])

        if p4_password:
            p4.password = p4_password
            p4.run_login()
            self.p4_connection = True

        self.p4 = p4

class PerforceClient:
    """
    Wrapper around the P4 workspace (client)
    """

    def __init__(self, connection: PerforceConnection, name: str, specification: dict = None):
        self.p4 = connection.p4
        self.name = name
        self.spec = self.p4.fetch_client(name)
        if specification:
            self.spec.update(specification)

    def save(self):
        """
        Save the perforce client (workspace) on the P4 server
        """

        self.p4.save_client(self.spec)

    def sync(self, filepath: str = None):
        """
        Execute `p4 sync` on the given file path. If no path given, will sync the whole workspace

        :param filepath: File path to sync
        """

        self.p4.client = self.name
        if filepath:
            self.p4.run('sync', filepath)
        else:
            self.p4.run('sync')


def get_perforce_workspace_specification(port: str = None, user: str = None, client: str = None) -> Optional[dict]:
    """
    Get perforce workspace specification using provided port, user and client.
    If some of the parameters are missing, defaults will be used

    :param port: P4 server address
    :param user: P4 user name
    :param client: P4 client (workspace) name

    :return: P4 workspace specification dictionary if successful, None otherwise
    :rtype: Optional[dict]
    """

    p4 = PerforceConnection(port=port, user=user).p4
    if client:
        p4.client = client

    try:
        workspace_specification = p4.fetch_client(p4.client)
        return workspace_specification
    except P4Exception as e:
        logger.info(str(e))


if __name__ == '__main__':
    p4 = PerforceConnection().p4
    print(p4.user)



#  Copyright 2024 CoreWeave
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import re
import unreal
import logging
from pathlib import Path


content_dir = unreal.Paths.convert_relative_path_to_full(unreal.Paths.project_content_dir())
saved_dir = unreal.Paths.convert_relative_path_to_full(unreal.Paths.project_saved_dir()).rstrip('/')


def get_project_file_path() -> str:
    """
    Returns the Unreal project OS path

    :return: the Unreal project OS path
    :rtype: str
    """

    if unreal.Paths.is_project_file_path_set():
        project_file_path = unreal.Paths.convert_relative_path_to_full(
            unreal.Paths.get_project_file_path()
        )
        return project_file_path
    else:
        raise RuntimeError(
            "Failed to get a project name. Please set a project!"
        )


def get_project_directory() -> str:
    """
    Returns the Unreal project directory OS path

    :return: the Unreal project directory OS path
    :rtype: str
    """

    project_file_path = get_project_file_path()
    project_directory = str(Path(project_file_path).parent).replace('\\', '/')
    return project_directory


def get_project_name() -> str:
    """
    Returns the Unreal project name without extension

    :return: Unreal project name
    :rtype: str
    """

    return Path(get_project_file_path()).stem


def soft_obj_path_to_str(soft_obj_path: unreal.SoftObjectPath) -> str:
    """
    Converts the given unreal.SoftObjectPath to the Unreal path

    :param soft_obj_path: unreal.SoftObjectPath instance

    :return: the Unreal path, e.g. /Game/Path/To/Asset
    :rtype: str
    """

    obj_ref = unreal.SystemLibrary.conv_soft_obj_path_to_soft_obj_ref(soft_obj_path)
    return unreal.SystemLibrary.conv_soft_object_reference_to_string(obj_ref)


def os_path_from_unreal_path(unreal_path, with_ext: bool = False):
    """
    Convert Unreal path to OS path, e.g. /Game/Assets/MyAsset to C:/UE_project/Content/Assets/MyAsset.uasset.

    if parameter with_ext is set to True, tries to get type of the asset by unreal.AssetData and set appropriate extension:

    - type World - .umap
    - other types - .uasset

    If for some reason it can't find asset data (e.g. temporary level's actors don't have asset data), it will set ".uasset"

    :param unreal_path: Unreal Path of the asset, e.g. /Game/Assets/MyAsset
    :param with_ext: if True, build the path with extension (.uasset or .umap), set asterisk "*" otherwise.

    :return: the OS path of the asset
    :rtype: str
    """

    os_path = str(unreal_path).replace('/Game/', content_dir)

    if with_ext:
        asset_data = unreal.EditorAssetLibrary.find_asset_data(unreal_path)
        asset_class_name = asset_data.asset_class_path.asset_name \
            if hasattr(asset_data, 'asset_class_path') \
            else asset_data.asset_class  # support older version of UE python API

        if not asset_class_name.is_none():  # AssetData not found - asset not in the project / on disk
            os_path += '.umap' if asset_class_name == 'World' else '.uasset'
        else:
            os_path += '.uasset'
    else:
        os_path += '.*'

    return os_path


def get_engine_version_number() -> str:
    """
    Get the engine version number in format `major.minor.patch`, for example `5.3.2`

    :return: Unreal Engine version number as string
    :rtype: str
    """

    return re.findall(r'\d\.\d\.\d', unreal.SystemLibrary.get_engine_version())[0]

def log(message: str, level: int = logging.INFO):
    """
    Log a message to the Unreal log with the specified level.

    :param message: The message to log.
    :param level: The logging library log level.
    :return: Formatted log message.
    :rtype: str
    """

    if level == logging.DEBUG:
        unreal.log(f"Debug: {message}")
    elif level == logging.INFO:
        unreal.log(message)
    elif level == logging.WARNING:
        unreal.log_warning(message)
    elif level == logging.ERROR:
        unreal.log_error(message)
    elif level == logging.CRITICAL:
        unreal.log_error(f"Critical: {message}")
    else:
        raise ValueError(f"Unreal Utils: Unsupported log level: {level}")


#  Copyright 2024 CoreWeave
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import random
import unittest


from ciounreal.common import conductor_data


class TestConductorData(unittest.TestCase):

    def test_get_ciocore_data(self):
        ciocore_data = conductor_data.ConductorData.get_ciocore_data()
        assert isinstance(ciocore_data, dict)

    def test_get_projects(self):
        projects = conductor_data.ConductorData.get_projects()
        assert isinstance(projects, list)
        assert len(projects) > 0

    def test_get_software(self):
        software = conductor_data.ConductorData.get_software()
        assert isinstance(software, conductor_data.PackageTree)

    def test_get_supported_software_packages(self):
        supported_software_packages = conductor_data.ConductorData.get_supported_software_packages()
        for package in supported_software_packages:
            assert package['platform'] == 'windows'
            assert package['product'] == 'unrealengine'

    def test_get_instance_types(self):
        isinstance_types = conductor_data.ConductorData.get_instance_types()
        for instance_type in isinstance_types:
            assert instance_type.get('gpu') is not None
            assert instance_type['operating_system'] == 'windows'

    def test_get_instance_type_by_description(self):
        instance_types = conductor_data.ConductorData.get_instance_types()
        random_index = random.randint(0, len(instance_types) - 1)
        instance_type_description = instance_types[random_index]['description']

        instance_type = conductor_data.ConductorData.get_instance_type_by_description(instance_type_description)
        assert instance_type in instance_types



#  Copyright 2024 CoreWeave
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import sys
import unreal
from ciounreal.common import unreal_utils

@unreal.uclass()
class MoviePipelineConductorLocalEditorExecutor(unreal.MoviePipelinePIEExecutor):
    """
    Movie Pipeline Executor that executes render and log progress to the console
    """

    totalFrameRange = unreal.uproperty(int)
    currentFrame = unreal.uproperty(int)

    def _post_init(self):
        """
        Constructor that gets called when created either via C++ or Python
        Note that this is different from the standard __init__ function of Python
        """

        self.totalFrameRange = 0
        self.currentFrame = 0

    @unreal.ufunction(override=True)
    def execute(self, queue: unreal.MoviePipelineQueue):
        """
        Execute the provided Queue.
        You are responsible for deciding how to handle each job in the queue and processing them.

        Here we define totalFrameRange as frames count from the sequence/job configuration

        :param queue: The queue that this should process all jobs for
        """

        try:
            # get the single job from queue
            jobs = queue.get_jobs()
            if len(jobs) == 0:
                self.log(f'Render Executor: Error: {queue} has {len(jobs)} jobs')

            job = jobs[0]

            # get output settings block
            output_settings = job.get_configuration().find_or_add_setting_by_class(
                unreal.MoviePipelineOutputSetting
            )

            # if user override frame range, use overridden values
            if output_settings.use_custom_playback_range:
                self.totalFrameRange = output_settings.custom_end_frame - output_settings.custom_start_frame

            # else use default frame range of the level sequence
            else:
                level_sequence = unreal.EditorAssetLibrary.load_asset(
                    unreal.SystemLibrary.conv_soft_object_reference_to_string(
                        unreal.SystemLibrary.conv_soft_obj_path_to_soft_obj_ref(
                            job.sequence
                        )
                    )
                )
                if level_sequence is None:
                    self.log('Error: Level Sequence not loaded. Check if the sequence exists and is valid')

                self.totalFrameRange = level_sequence.get_playback_end() - level_sequence.get_playback_start()

            if self.totalFrameRange == 0:
                self.log(f'Cannot render the Queue with frame range of zero length')
        except Exception as e:
            exit_code = 102
            self.log(f'Error handled during rendering: {str(e)}. Quit with Exit Code {exit_code}')
            unreal.ConductorMiscFunctionLibrary.request_exit(True, exit_code)

        # don't forget to call parent's execute to run the render process
        super().execute(queue)

    @unreal.ufunction(override=True)
    def on_begin_frame(self):
        """
        Called once at the beginning of each engine frame (e.g. tick, fps)
        Since the executor will work with Play in Editor widget, each rendered frame will match with widget frame tick.
        """

        super(MoviePipelineConductorLocalEditorExecutor, self).on_begin_frame()

        # Since PIEExecutor launching Play in Editor before mrq is rendering, we should ensure, that
        # executor actually rendering the sequence.
        if self.is_rendering():
            self.currentFrame += 1
            progress = self.currentFrame / self.totalFrameRange * 100

            # Executor work with the render queue after all frames are rendered - do all
            # support stuff, handle safe quit, etc, so we should ignore progress that more than 100
            if progress <= 100:
                self.log(f'Progress: {progress}')

    @classmethod
    def log(cls, message: str):
        """
        Log message to the console with prepended executor name

        :param message: message to log
        """

        unreal_utils.log('MoviePipelineConductorLocalEditorExecutor: {}'.format(message))



#  Copyright 2024 CoreWeave
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import sys
import time
import unreal
import unittest
from unittest.mock import Mock, MagicMock, PropertyMock, patch

from ciounreal.conductor_job_submitter import ConductorJobSubmitter, UnrealConductorJobSubmitStatus


PIPELINE_QUEUE = unreal.get_editor_subsystem(unreal.MoviePipelineQueueSubsystem).get_queue()


def submission_main_mock():
    return {'body': 'job submitted.', 'jobid': '00004', 'status': 'success', 'uri': '/jobs/00004'}, 201


def submission_main_fail_mock():
    return {'body': 'Frightful error', 'status': 'error'}, 500


class TestUnrealSubmitter(unittest.TestCase):

    def test_add_job(self, submitter: ConductorJobSubmitter = None, delete: bool = False):

        submitter = submitter or ConductorJobSubmitter(silent_mode=True)

        for job in PIPELINE_QUEUE.get_jobs():
            submitter.add_job(job)

        self.assertIsNot(len(submitter._jobs), 0)

        if delete:
            submitter.delete_jobs()

    @patch('ciounreal.conductor_job_submitter.Submit.main', side_effect=submission_main_mock)
    def test_submit_jobs(self, submission_main_mock: Mock):

        submitter = ConductorJobSubmitter(silent_mode=True)
        self.test_add_job(submitter, delete=False)
        submitter.submit_jobs()

        submission_main_mock.assert_called_once()

    @patch('ciounreal.conductor_job_submitter.Submit.main', side_effect=submission_main_mock)
    @patch('ciounreal.conductor_job_submitter.ConductorJobSubmitter._submission_result_dialog', new_callable=MagicMock)
    def test_cancel_submit_jobs(self, submission_main_mock: Mock, submission_result_dialog_mock: MagicMock):

        submitter = ConductorJobSubmitter(silent_mode=True)
        self.test_add_job(submitter, delete=False)

        with patch.object(submitter, '_submit_status', UnrealConductorJobSubmitStatus.CANCELED):
            submitter.submit_jobs()

        submission_main_mock.assert_called_once()

        submission_result_dialog_mock.assert_called_once_with(
            f'Jobs submission canceled.\n'
            f'Number of not submitted jobs: 1'
        )

    @patch('ciounreal.conductor_job_submitter.ConductorJobSubmitter.submit_failed_message', new_callable=PropertyMock)
    @patch('ciounreal.conductor_job_submitter.Submit.main', side_effect=submission_main_fail_mock)
    @patch('ciounreal.conductor_job_submitter.ConductorJobSubmitter._submission_result_dialog', new_callable=MagicMock)
    def test_fail_submit_jobs(
            self,
            submission_result_dialog_mock: MagicMock,
            submission_main_fail_mock: Mock,
            submit_failed_message_mock: Mock
    ):

        submit_failed_message_mock.side_effect = ['Test interrupt submission']

        submitter = ConductorJobSubmitter(silent_mode=True)
        self.test_add_job(submitter, delete=False)

        with patch.object(submitter, '_submit_status', UnrealConductorJobSubmitStatus.FAILED):
            submitter.submit_jobs()

        submission_main_fail_mock.assert_called_once()

        submit_failed_message_mock.assert_called_once()


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(TestUnrealSubmitter)
    unittest.TextTestRunner(stream=sys.stdout, buffer=True).run(suite)



#  Copyright 2024 CoreWeave
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import unreal
import logging

from ciounreal.common import unreal_utils
from ciounreal.conductor_job_submitter import ConductorJobSubmitter


@unreal.uclass()
class MoviePipelineConductorRemoteExecutor(unreal.MoviePipelineExecutorBase):
    """
    Movie Pipeline Executor that submit jobs to the Conductor
    """

    # The queue we are working on, null if no queue has been provided.
    pipeline_queue = unreal.uproperty(unreal.MoviePipelineQueue)
    job_ids = unreal.uproperty(unreal.Array(str))

    @unreal.ufunction(override=True)
    def execute(self, pipeline_queue):
        """
        Execute the provided Queue.
        You are responsible for deciding how to handle each job in the queue and processing them.

        Here we define totalFrameRange as frames count from the sequence/job configuration

        :param queue: The queue that this should process all jobs for
        :return: None
        """

        unreal_utils.log(f"Asked to execute Queue: {pipeline_queue}")
        unreal_utils.log(f"Queue has {len(pipeline_queue.get_jobs())} jobs")

        if not pipeline_queue or (not pipeline_queue.get_jobs()):
            self.on_executor_finished_impl()
            return

        if not self.valid_shots(pipeline_queue):
            return
    
        if not self.check_dirty_packages():
            return

        if not self.check_maps(pipeline_queue):
            return

        self.pipeline_queue = pipeline_queue

        self._submit_jobs()

    def _submit_jobs(self):
        """
        Submit valid, enabled jobs to Conductor.
        """
        submitter = ConductorJobSubmitter()
        
        empty_jobs = self._invalid_shots(self.pipeline_queue)
        enabled_jobs = [job for job in self.pipeline_queue.get_jobs() if job.is_enabled()]

        valid_jobs = list(set(enabled_jobs) - set(empty_jobs))

        for job in valid_jobs:
            unreal_utils.log(f"Submitting Job `{job.job_name}` to CoreWeave Conductor ...")
            submitter.add_job(job)
        
        submitter.submit_jobs()

    @unreal.ufunction(override=True)
    def is_rendering(self):
        """
        Because we forward unfinished jobs onto another service when the
        button is pressed, they can always submit what is in the queue and
        there's no need to block the queue.

        A MoviePipelineExecutor implementation must override this. If you
        override a ufunction from a base class you don't specify the return
        type or parameter types

        :return: False
        :rtype: bool
        """

        return False

    def check_dirty_packages(self) -> bool:
        """
        Checks if the current project has dirty unsaved packages

        :return: True if there is unsaved content, False otherwise
        :rtype: bool
        """

        dirty_packages = []
        dirty_packages.extend(
            unreal.EditorLoadingAndSavingUtils.get_dirty_content_packages()
        )
        dirty_packages.extend(
            unreal.EditorLoadingAndSavingUtils.get_dirty_map_packages()
        )

        if dirty_packages:
            if not unreal.EditorLoadingAndSavingUtils.save_dirty_packages_with_dialog(
                True, True
            ):
                message = (
                    "One or more jobs in the queue have an unsaved map/content. "
                    "{packages} "
                    "Please save and check-in all work before submission.".format(
                        packages="\n".join(dirty_packages)
                    )
                )

                unreal_utils.log(message, logging.ERROR)
                unreal.EditorDialog.show_message(
                    "Unsaved Maps/Content", message, unreal.AppMsgType.OK
                )
                self.on_executor_finished_impl()
                return False
        return True

    def valid_shots(self, pipeline_queue) -> bool:
        """
        Checks if the jobs have at least one valid shot to render

        :return: True if valid, False otherwise
        :rtype: bool
        """

        empty_jobs = self._invalid_shots(pipeline_queue, job_names=True)
        if empty_jobs:
            message = (
                "Would you like to continue this submission?\n\n" +
                "The following job(s) will be skipped because they do not contain any enabled shots: " +
                "{}. ".format(", ".join(empty_jobs))
                
            )
            unreal_utils.log(message, logging.ERROR)
            result = unreal.EditorDialog.show_message(
                "Empty Jobs", message, unreal.AppMsgType.OK_CANCEL
            )

            if result == unreal.AppReturnType.CANCEL:
                self.on_executor_finished_impl()
                return False

        return True 
    
    def _invalid_shots(self, pipeline_queue, job_names=False) -> list:
        """
        Checks if the jobs have at least one valid shot to render

        :return: list of jobs with no enabled shots
        :rtype: list(unreal.MoviePipelineExecutorJob)
        """
        jobs = pipeline_queue.get_jobs()
        empty_jobs = []

        for i in range(len(jobs)):
            job = jobs[i]
            shots = job.shot_info
            disabled_shots = [shots[i] for i in range(len(shots)) if not shots[i].enabled ]
           
            if len(disabled_shots) == len(shots):
                if job_names:
                    empty_jobs.append(job.job_name)
                else:
                    empty_jobs.append(job)
        
        return empty_jobs
    
    def check_maps(self, pipeline_queue) -> bool:
        """
        Checks if the jobs' maps is valid for remote render
        (see `unreal.MoviePipelineEditorLibrary.is_map_valid_for_remote_render <https://dev.epicgames.com/documentation/en-us/unreal-engine/python-api/class/MoviePipelineEditorLibrary?application_version=5.3#unreal.MoviePipelineEditorLibrary.is_map_valid_for_remote_render>`__)


        :return: True if valid, False otherwise
        :rtype: bool
        """

        has_valid_map = (
            unreal.MoviePipelineEditorLibrary.is_map_valid_for_remote_render(
                pipeline_queue.get_jobs()
            )
        )
        if not has_valid_map:
            message = (
                "One or more jobs in the queue have an unsaved map as "
                "their target map. "
                "These unsaved maps cannot be loaded by an external process, "
                "and the render has been aborted."
            )
            unreal_utils.log(message, logging.ERROR)
            unreal.EditorDialog.show_message(
                "Unsaved Maps", message, unreal.AppMsgType.OK
            )
            self.on_executor_finished_impl()
            return False

        return True



#  Copyright 2024 CoreWeave
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import os
import unreal
import pprint

from ciounreal.common import context
from ciounreal.common import unreal_utils
from ciounreal.conductor_job import ConductorJob, PerforceConductorJob


class ConductorJobBuilder:

    """
    Helper class to build a Conductor Job
    """

    @classmethod
    def get_build_context(cls, mrq_job: unreal.ConductorMoviePipelineExecutorJob) -> context.Context:
        """
        Get build context from the given unreal.ConductorMoviePipelineExecutorJob

        :param mrq_job: unreal.ConductorMoviePipelineExecutorJob
        :return: :class:`ciounreal.common.context.Context` object
        :rtype: :class:`ciounreal.common.context.Context`
        """

        level_sequence_path = os.path.splitext(unreal_utils.soft_obj_path_to_str(mrq_job.sequence))[0]
        level_sequence_name = level_sequence_path.split("/")[-1]

        map_path = os.path.splitext(unreal_utils.soft_obj_path_to_str(mrq_job.map))[0]
        map_name = level_sequence_path.split("/")[-1]

        output_settings = mrq_job.get_configuration().find_setting_by_class(unreal.MoviePipelineOutputSetting)

        build_context = context.Context(
            {
                'project_path': unreal_utils.get_project_file_path(),
                'project_dir': unreal_utils.get_project_directory(),
                'job_name': mrq_job.job_name,
                'level_sequence': level_sequence_path,
                'level_sequence_name': level_sequence_name,
                'sequence_name': level_sequence_name,
                'map_path': map_path,
                'map_name': map_name,
                'level_name': map_name,
                'resolution': f'{output_settings.output_resolution.x}x{output_settings.output_resolution.y}',
            }
        )
        build_context.update(
            {
                'output_dir': output_settings.output_directory.path.format_map(build_context).replace(
                    "\\", "/"
                ).rstrip("/"),
                'filename_format': output_settings.file_name_format.format_map(build_context)
            }
        )

        return build_context

    def build_conductor_job(self, mrq_job: unreal.ConductorMoviePipelineExecutorJob) -> ConductorJob:
        """
        Build a Conductor Job from the given unreal.ConductorMoviePipelineExecutorJob

        :param mrq_job: unreal.ConductorMoviePipelineExecutorJob

        :return: ConductorJob object
        :rtype: Union[:class:`ciounreal.conductor_job.conductor_job.ConductorJob`,
                      :class:`ciounreal.conductor_job.perforce_conductor_job.PerforceConductorJob`]
        """

        build_context = self.get_build_context(mrq_job)
        job_class = PerforceConductorJob if mrq_job.conductor_settings.perforce_settings.use_perforce else ConductorJob
        job = job_class(mrq_job, build_context=build_context)
        unreal_utils.log(f'Conductor Job:')
        unreal_utils.log(pprint.pformat(job.as_dict()))
        return job



# Copyright 2024 CONDUCTOR TECHNOLOGIES. All Rights Reserved.

import os
import sys

import unreal
from ciounreal.common import unreal_utils

unreal_utils.log('Initialize Conductor plugin')

from p4utilsforunreal import perforce
from ciounreal.common import conductor_data
from ciounreal.render_executors.remote_render_executor import MoviePipelineConductorRemoteExecutor
from ciounreal.render_executors.local_editor_render_executor import MoviePipelineConductorLocalEditorExecutor


os.environ['UE_CONDUCTOR_PLUGIN_PATH'] = os.path.dirname(
    os.path.dirname(
        os.path.dirname(
            __file__
        )
    )
).replace('\\', '/')
 
if os.environ.get("CONDUCTOR_DISABLE_PLUGIN") == '1':
    unreal.log("Skipping optional Conductor plugin login.")
else: 
    conductor_data.ConductorData.get_ciocore_data() # init plugin data

def refresh_conductor_data():
    with unreal.ScopedSlowTask(1, "Refresh Conductor data") as slow_task:
        slow_task.make_dialog(True)
        conductor_data.ConductorData.get_ciocore_data()
        slow_task.enter_progress_frame(1, "Refresh Conductor data")

@unreal.uclass()
class ConductorSettingsLibraryImplementation(unreal.ConductorSettingsLibrary):

    """
    Python implementation of the unreal.ConductorSettingsLibrary C++ class
    which perform retrieving settings options for the Conductor Job
    """

    @unreal.ufunction(override=True)
    def get_job_title(self):
        """
        Return Conductor Job title. By default equals to `{job_name}-{level_sequence}`

        :return: Conductor Job title as string
        :rtype: str
        """

        return "{job_name}-{level_sequence}"

    @unreal.ufunction(override=True)
    def get_instance_types(self):
        return [instance_type['description'] for instance_type in conductor_data.ConductorData.get_instance_types()]

    @unreal.ufunction(override=True)
    def get_env_merge_policy(self):
        return ["append", "prepend", "exclusive"]

    @unreal.ufunction(override=True)
    def get_projects(self):
        return conductor_data.ConductorData.get_projects()

    @unreal.ufunction(override=True)
    def get_default_task_template(self):
        return 'UnrealEditor-Cmd.exe "{project_path}" ' \
               '-windowed -log -stdout -FullStdOutLogOutput -unattended ' \
               '-RenderOffscreen -noloadingscreen -allowstdoutlogverbosity ' \
               '-QueueManifest="{queue_manifest}" ' \
               '-LevelSequence="{level_sequence}" ' \
               '-Level="{map_path}" ' \
               '-execcmds="r.HLOD 0, py mrq_cli.py"'

    @unreal.ufunction(override=True)
    def get_perforce_server(self):
        return perforce.PerforceConnection().p4.port
    
    @unreal.ufunction(override=True)
    def get_perforce_username(self):
        return perforce.PerforceConnection().p4.user

    @unreal.ufunction(override=True)
    def reconnect(self):
        refresh_conductor_data()


def add_toolbar_menu():
    command = 'refresh_conductor_data()'

    menus = unreal.ToolMenus.get()
    toolbar = menus.find_menu('LevelEditor.LevelEditorToolBar.PlayToolBar')
    toolbar.add_section(section_name='Conductor', label='Conductor')

    entry = unreal.ToolMenuEntry(type=unreal.MultiBlockType.TOOL_BAR_BUTTON)
    entry.set_label('Conductor')
    entry.set_tool_tip('Refresh conductor data')
    entry.set_icon('ConductorStyle', 'Conductor.Icon')
    entry.set_string_command(
        type=unreal.ToolMenuStringCommandType.PYTHON,
        custom_type=unreal.Name(''),
        string=command
    )
    toolbar.add_menu_entry('Conductor', entry)
    menus.refresh_all_widgets()


unreal_utils.log("Add Conductor toolbar menu")
add_toolbar_menu()



#  Copyright 2024 CoreWeave
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import os
import glob
import unreal
import unittest


from ciounreal.common import unreal_utils


class TestUnrealUtils(unittest.TestCase):

    @classmethod
    def create_test_asset(cls) -> str:
        factory = unreal.BlueprintFactory()
        factory.set_editor_property("ParentClass", unreal.Actor)

        asset_name = 'TestAsset'
        asset_package_path = '/Game'
        asset_path = f'{asset_package_path}/{asset_name}'

        asset_tools = unreal.AssetToolsHelpers.get_asset_tools()
        asset_tools.create_asset(
            asset_name=asset_name,
            package_path=asset_package_path,
            asset_class=None,
            factory=factory
        )
        unreal.EditorAssetLibrary.save_asset(asset_path)

        return asset_path

    def test_get_project_file_path(self):
        project_file_path = unreal_utils.get_project_file_path()
        assert os.path.exists(project_file_path)

    def test_get_project_directory(self):
        project_directory = unreal_utils.get_project_directory()
        assert os.path.exists(project_directory)

    def test_soft_object_path_to_str(self):
        original_path = '/Game/Assets/TestAsset'
        soft_object_path = unreal.SoftObjectPath(original_path)
        assert original_path == unreal_utils.soft_obj_path_to_str(soft_object_path)

    def test_os_path_from_unreal_path(self):
        asset_path = self.create_test_asset()

        os_path_with_ext = unreal_utils.os_path_from_unreal_path(asset_path, with_ext=True)
        assert os.path.exists(os_path_with_ext)

        os_path_without_ext = unreal_utils.os_path_from_unreal_path(asset_path, with_ext=False)
        assert len(glob.glob(os_path_without_ext)) > 0

        unreal.EditorAssetLibrary.delete_asset(asset_path)

    def test_get_engine_version_number(self):
        engine_version = unreal.SystemLibrary.get_engine_version()
        engine_version_number = unreal_utils.get_engine_version_number()

        assert engine_version.startswith(engine_version_number)

        major_variants = [4, 5]
        minor_variants = [v for v in range(0, 10)]
        patch_variants = [v for v in range(0, 10)]

        major, minor, patch = engine_version_number.split('.')
        assert int(major) in major_variants
        assert int(minor) in minor_variants
        assert int(patch) in patch_variants










#  Copyright 2024 CoreWeave
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import os
import json
import unreal
import logging
import datetime
import threading
import traceback
from typing import Union
from enum import IntEnum

from ciocore import config
from ciocore.conductor_submit import Submit
from ciocore.uploader.upload_stats import UploadStats

from ciounreal.common import unreal_utils
from ciounreal.conductor_job import ConductorJob
from ciounreal.conductor_job_builder import ConductorJobBuilder


class UnrealConductorJobSubmitStatus(IntEnum):
    """
    Job submission status
    """

    UPLOADING = 1
    COMPLETED = 2
    FAILED = 3
    CANCELED = 4


# TODO make pretty clickable link
class SubmittedConductorJobDescriptor:
    """
    Submitted Conductor Job payload.
    Consist of job name, https response and response code
    """

    def __init__(self, job_name: str, submission_response: dict, submission_response_code: int):
        self.job_name = job_name
        self.submission_response = submission_response
        self.submission_response_code = submission_response_code

        if isinstance(self.submission_response_code, int) and self.submission_response_code <= 201:
            job_uri = self.submission_response['uri'].rstrip('/').lstrip('/').replace('jobs', 'job')
        else:
            job_uri = ''

        self.job_url = f'{config.config().config["url"]}/{job_uri}'

    def __str__(self):
        return f'{self.job_name} - {self.job_url}'


class ConductorJobSubmitter:
    """
    Main class that manage Conductor Job submission process
    """

    def __init__(self, silent_mode: bool = False):
        self._silent_mode = silent_mode

        self._jobs: list[ConductorJob] = []
        self._job_builder = ConductorJobBuilder()

        self._submission: Union[Submit, None] = None
        self._submit_status: Union[UnrealConductorJobSubmitStatus, None] = None
        self._submit_message: Union[str, None] = None
        self._submit_failed_message: Union[str, None] = None

        self._progress_list: list[float] = []
        self._submitted_job_descriptors: list[SubmittedConductorJobDescriptor] = []

    @property
    def submit_failed_message(self):
        """
        Last submission error message

        :return: Last submission error message
        :rtype: str
        """

        return self._submit_failed_message

    @classmethod
    def save_conductor_job_to_file(cls, job: ConductorJob) -> str:
        """
        .. warning::
           FOR DEBUG ONLY

        Save Conductor Job payload to JSON file to %APPDATA%/Conductor/ciounreal/jobs/<current_datetime>.json
        instead of real submit

        To use it, set env variable CIOUNREAL_DEBUG as "True":
        os.environ['CIOUNREAL_DEBUG'] = 'True'

        To not, set env variable CIOUNREAL_DEBUG as "False":
        os.environ['CIOUNREAL_DEBUG'] = 'False'

        :param job: :class:`ciounreal.conductor_job.conductor_job.ConductorJob`
        :return: path to saved JSON file
        :rtype: str
        """

        jobs_directory = f'{os.getenv("APPDATA")}/Conductor/ciounreal/jobs'
        os.makedirs(jobs_directory, exist_ok=True)
        with open(f'{jobs_directory}/{datetime.datetime.now().strftime("%m-%d-%Y-%H-%M-%S")}.json', 'w') as f:
            json.dump(
                job.as_dict(),
                f,
                indent=4
            )

        return f.name

    def _display_progress(self, submit_status: UnrealConductorJobSubmitStatus, title: str):
        """
        Display submission progress with unreal.ScopedSlowTask

        :param submit_status: target submission status
        :param title: target title to display
        """

        last_progress = 0
        with unreal.ScopedSlowTask(100, title) as submit_task:
            submit_task.make_dialog(can_cancel=False)  # Cancel work only for uploader, so disable cancelling here
            while self._submit_status == submit_status:
                if self._submit_status == UnrealConductorJobSubmitStatus.FAILED:
                    break

                if submit_task.should_cancel():
                    self._submission.stop_work()
                    self._submit_status = UnrealConductorJobSubmitStatus.CANCELED
                    break

                if len(self._progress_list) > 0:
                    new_progress = self._progress_list.pop(0)
                else:
                    new_progress = last_progress
                # work_increment = new_progress - last_progress if new_progress > last_progress else 0
                # submit_task.enter_progress_frame(work_increment, self._submit_message)
                submit_task.enter_progress_frame(new_progress - last_progress, self._submit_message)
                last_progress = new_progress

    def _uploader_time_formatter(self, time_delta):
        """
        Format uploader time (datetime.timedelta) to string.
        
        :param time_delta: Uploader time ('N/A', 'hh:mm:ss.sss', 'hh:mm:ss')
        :return: Formatted upload time
        :rtype: str
        """
        return f'{time_delta}'.split('.')[0]
    
    def _uploader_progress_handler(self, upload_stats: UploadStats):
        """
        Callback function that passed to :class:`ciocore.conductor_submit.Submit`
        to track current upload progress and display in UI

        :param upload_stats: :class:`ciocore.uploader.upload_stats.UploadStats`
        """

        self._submit_status = UnrealConductorJobSubmitStatus.UPLOADING

        unreal_utils.log(f'Uploader progress: {upload_stats.get_formatted_text()}')

        elapsed_time = self._uploader_time_formatter(upload_stats.elapsed_time)
        time_remaining = self._uploader_time_formatter(upload_stats.time_remaining)

        self._submit_message = f'{upload_stats.bytes_uploaded} / {upload_stats.bytes_to_upload}, ' \
                               f'transfer rate {upload_stats.transfer_rate}/s, ' \
                               f'{elapsed_time} / {time_remaining}'

        completion = upload_stats.percent_complete.value  # Can be N/A
        if completion:
            self._progress_list.append(completion * 100)  # value between 0.0 and 1.0, convert to UE 0.0 to 100.0

    def _start_submit(self, job: ConductorJob):
        """
        Start submission process of the given job

        :param job: :class:`ciounreal.conductor_job.conductor_job.ConductorJob`
        """
        try:
            self._submission = Submit(job.as_dict())
            if job.local_upload:
                self._submission.progress_handler = self._uploader_progress_handler
            response, response_code = self._submission.main()
            unreal_utils.log(f'Submission response: {response}, {response_code}')

            self._submit_status = UnrealConductorJobSubmitStatus.COMPLETED
            self._submitted_job_descriptors.append(
                SubmittedConductorJobDescriptor(job.job_title, response, response_code)
            )

        except Exception as e:
            unreal_utils.log(str(e), logging.ERROR)
            self._submit_failed_message = str(e)
            self._submit_status = UnrealConductorJobSubmitStatus.FAILED

    def _submission_result_dialog(
            self,
            message: str,
            title='Conductor Job Submission',
            message_type=unreal.AppMsgType.OK
    ):
        """
        Display submission result dialog

        :param message: Submission result message
        :param title: Submission result title
        :param message_type: unreal message type (see unreal.AppMsgType)
        """

        unreal_utils.log(f'{title}: {message}')

        if self._silent_mode:
            return

        unreal.EditorDialog.show_message(
            title=unreal.Text(title),
            message=unreal.Text(message),
            message_type=message_type
        )

    def delete_jobs(self):
        """
        Clear list of the jos to submit
        """

        self._jobs.clear()

    def add_job(self, mrq_job: unreal.ConductorMoviePipelineExecutorJob):
        """
        Add a new job to the submission list

        :param mrq_job: unreal.ConductorMoviePipelineExecutorJob
        """

        try:
            self._jobs.append(self._job_builder.build_conductor_job(mrq_job))
        except Exception as e:
            self._submission_result_dialog(f'Job {mrq_job.job_name} failed to submit due to an error: \n{str(e)}')
            unreal_utils.log(traceback.format_exc())

    def submit_jobs(self):
        """
        Submit all the added jobs to Conductor.
        """

        for job in self._jobs:
            job._save_queue_manifest_file()

            if os.getenv('CIOUNREAL_DEBUG', 'False') == 'True':
                save_path = self.save_conductor_job_to_file(job)
                self._submission_result_dialog(
                    message=f'Job saved in {save_path}',
                    title='Conductor Job Submission DEBUG'
                )
                continue

            self._submit_status = UnrealConductorJobSubmitStatus.UPLOADING
            self._submit_message = 'Start submitting ...'
            self._progress_list = []

            t = threading.Thread(target=self._start_submit, args=(job,), daemon=True)
            t.start()

            self._display_progress(submit_status=UnrealConductorJobSubmitStatus.UPLOADING, title='Uploading')

            t.join()

            unreal_utils.log(f'SUBMISSION STATUS: {self._submit_status.value}')
            if self._submit_status == UnrealConductorJobSubmitStatus.FAILED:
                self._submission_result_dialog(
                    f'Job {job.job_title} failed to submit due to an error: \n{self.submit_failed_message}'
                )

            elif self._submit_status == UnrealConductorJobSubmitStatus.CANCELED:
                self._submission_result_dialog(
                    f'Jobs submission canceled.\n'
                    f'Number of not submitted jobs: {len(self._jobs) - len(self._submitted_job_descriptors)}'
                )
                break

        self._submission_result_dialog(
            f'Submitted jobs ({len(self._submitted_job_descriptors)}):\n' +
            '\n'.join([str(j) for j in self._submitted_job_descriptors])
        )



#  Copyright 2024 CoreWeave
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import unreal
from typing import Callable
from dataclasses import dataclass, asdict

from ciounreal.common import unreal_utils


asset_registry = unreal.AssetRegistryHelpers.get_asset_registry()


class DependencyFilters:
    """
    Class container that contains methods for filtering dependencies
    """

    @staticmethod
    def dependency_in_game_folder(dependency_path: str) -> bool:
        """Checks if /Game/ exists in passed path

        :param dependency_path: The path to dependency
        :return: True if dependency path in `/Game/`, False otherwise
        :rtype: bool
        """
        return '/Game/' in str(dependency_path)


@dataclass
class DependencyOptions:
    """
    Dataclass which contains the default options for getting the asset dependencies
    """

    include_hard_package_references: bool = True
    include_soft_package_references: bool = True
    include_hard_management_references: bool = True
    include_soft_management_references: bool = True
    include_searchable_names: bool = False

    def as_dict(self):
        """
        Return Dependency Search Options as dictionary

        :return: search options dictionary
        :rtype: dict
        """

        return asdict(self)


class DependencyCollector:

    def __init__(self):
        self._all_dependencies = list()
        self._missing_dependencies = list()
        self._already_synced = list()

    def collect(
            self,
            asset_path: str,
            dependency_options=DependencyOptions(),
            filter_method: Callable = None,
            sync_missing: bool = True
    ) -> list[str]:

        """The method starts the algorithm for obtaining dependencies on the passed asset

        :param asset_path: Path to the dependency asset to be obtained
        :param dependency_options: Settings for obtaining dependencies.
        :param filter_method: Method by which dependencies will be filtered for synchronization
        :param sync_missing: Whether to sync missing dependencies or not

        :return: list of collected dependencies (UE paths, start with /Game, etc.)
        :rtype: list[str]
        """

        self._all_dependencies.clear()

        udependency_options = unreal.AssetRegistryDependencyOptions(
            **dependency_options.as_dict()
        )

        if not unreal.EditorAssetLibrary.does_asset_exist(asset_path):
            self._sync_assets([asset_path])

        all_dependencies, missing_dependencies = self._get_dependencies(asset_path, udependency_options, filter_method)
        while missing_dependencies:
            if sync_missing:
                self._sync_assets(missing_dependencies)
            # Anyway extend synced even if source control is not available.
            # In that case first recursion before loop and recursion in loop will run one time
            self._already_synced.extend(missing_dependencies)
            self._missing_dependencies.clear()
            self._all_dependencies.clear()
            all_dependencies, missing_dependencies = self._get_dependencies(asset_path, udependency_options, filter_method)

        self._sync_assets(list(set([d for d in self._all_dependencies if d not in self._already_synced])))

        return list(set(self._all_dependencies))

    def _get_dependencies(
            self,
            asset_path: str,
            udependency_options: unreal.AssetRegistryDependencyOptions,
            filter_method: Callable = None
    ) -> tuple[list, list]:
        """
        The method recursively all dependencies on the passed asset

        :param asset_path: Path to the dependency asset to be obtained
        :param udependency_options: Settings for obtaining dependencies
        :param filter_method: Method by which dependencies will be filtered for synchronization

        :return: List of all downloaded dependencies and list of missing dependencies
        :rtype: list[str]
        """

        dependencies_raw = asset_registry.get_dependencies(
            package_name=asset_path,
            dependency_options=udependency_options
        )

        missing_dependencies = list()
        all_dependencies = list()
        if dependencies_raw:
            for dependency_raw in dependencies_raw:
                dependency_path = str(dependency_raw)
                does_confirm_filter = filter_method(dependency_path) if filter_method else True
                is_not_collected = dependency_raw not in self._all_dependencies

                if does_confirm_filter and is_not_collected:
                    # If Source Control off, last missed deps (synced or not) will be already in already synced list.
                    # So we don't fall in infinite recursion
                    is_missing = not unreal.EditorAssetLibrary.does_asset_exist(dependency_path) and dependency_path not in self._already_synced
                    missing_dependencies.append(dependency_path) if is_missing else all_dependencies.append(dependency_path)

        if all_dependencies:
            self._all_dependencies.extend(all_dependencies)

        if missing_dependencies:
            self._missing_dependencies.extend(missing_dependencies)

        for dependency in all_dependencies:
            self._get_dependencies(dependency, udependency_options, filter_method)

        return list(set(self._all_dependencies)), list(set(self._missing_dependencies))

    def _sync_assets(self, asset_paths: list[str]):
        """
        Sync given asset paths via `unreal.SourceControl <https://dev.epicgames.com/documentation/en-us/unreal-engine/python-api/class/SourceControl?application_version=5.3#unreal.SourceControl>`__

        :param asset_paths: List of assets to sync
        """

        if not asset_paths:
            return
        
        if not unreal.SourceControl.is_enabled():
            unreal.log('Source Control is not enabled. Skipping sync.')
            return

        synced = unreal.SourceControl.sync_files(
            [unreal_utils.os_path_from_unreal_path(path, with_ext=False) for path in asset_paths]
        )

        if not synced:
            unreal_utils.log(
                'Failed to complete the synchronization process. Reason: {}'.format(
                    unreal.SourceControl.last_error_msg()
                )
            )

        unreal.AssetRegistryHelpers().get_asset_registry().scan_modified_asset_files(asset_paths)
        unreal.AssetRegistryHelpers().get_asset_registry().scan_paths_synchronous(asset_paths, True, True)



#  Copyright 2024 CoreWeave
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import os
import re
import json
import unreal
from typing import Optional

from p4utilsforunreal import perforce

from ciounreal import exceptions
from ciounreal.common import context
from ciounreal.common import unreal_utils
from ciounreal.conductor_job import ConductorJob

from ciocore.package_environment import PackageEnvironment

class PerforceConductorJob(ConductorJob):

    def __init__(self, mrq_job: unreal.ConductorMoviePipelineExecutorJob = None, build_context: context.Context = None):
        if not mrq_job and not build_context:
            return
        
        self.perforce_settings = mrq_job.conductor_settings.perforce_settings
        perforce.PerforceConnection(port=self.perforce_settings.perforce_server,
                                    user=self.perforce_settings.perforce_username,
                                    password=self.perforce_settings.perforce_password,
                                    required_connection=True)

        self._perforce_workspace_specification_template = self.get_perforce_workspace_specification_template()
        self._perforce_workspace_specification_template_path = self._save_perforce_specification_template_file()

        super().__init__(mrq_job, build_context)

    @classmethod
    def get_unreal_python_path(cls):
        """
        Get the local OS path to the Unreal Engine Python executable

        :return: UE Python executable path, e.g.
                `C:/Program Files/Epic Games/UE_5.3/Engine/Binaries/ThirdParty/Python3/Win64/python.exe`
        :rtype: str
        """

        path_value = next(
            (env['value'] for env in cls.get_current_software_package()['environment'] if env['name'] == 'Path'),
            None
        )
        path_value = path_value.replace('\\', '/')

        ue_binaries_directory = next(
            (p for p in path_value.split(';') if 'Engine/Binaries/Win64' in p),
            None
        )

        unreal_python_path = f'{os.path.dirname(ue_binaries_directory)}/' \
                             f'ThirdParty/Python3/Win64/python.exe'.replace('\\', '/')

        return unreal_python_path

    @classmethod
    def get_perforce_settings_from_ue_source_control(cls) -> tuple[str, str, str]:
        """
        Parse /Saved/Config/WindowsEditor/SourceControlSettings.ini file inside project's directory
        and return P4 Port, User and Workspace

        :raises: :class:`ciounreal.exceptions.SourceControlProviderError`: If source control is not available or provider is invalid
        :return: P4 Port, User and Workspace tuple
        :rtype: tuple[str, str, str]
        """

        if not unreal.SourceControl.is_available():
            raise exceptions.SourceControlProviderError('Unreal Source Control is not available')

        source_control_ini = f'{unreal_utils.get_project_directory()}' \
                             f'/Saved/Config/WindowsEditor/SourceControlSettings.ini'
        if os.path.exists(source_control_ini):
            with open(source_control_ini, 'r') as f:
                lines = f.readlines()
            provider = next((l.split('=')[1].replace('\n', '') for l in lines if l.startswith('Provider')))
            if provider != 'Perforce':
                raise exceptions.SourceControlProviderError(
                    'Perforce is not current Source Control provider. Please, update settings and reopen the project. '
                    f'Current provider: {provider}'
                )
            port = next((l.split('=')[1].replace('\n', '') for l in lines if l.startswith('Port')), None)
            user = next((l.split('=')[1].replace('\n', '') for l in lines if l.startswith('UserName')), None)
            workspace = next((l.split('=')[1].replace('\n', '') for l in lines if l.startswith('Workspace')), None)
            return port, user, workspace
        else:
            raise FileNotFoundError(f'SourceControlSettings.ini not found: {source_control_ini}')

    @classmethod
    def get_perforce_workspace_specification_template(cls) -> Optional[dict]:
        """
        Build and return Perforce workspace specification template (dictionary) that consist of

        - Client (workspace name)
        - Root (workspace local root)
        - Stream (connected Stream name)
        - View (list of view path mappings, if no Stream was defined in original configuration - non-stream workspace)

        :raises: :class:`ciounreal.exceptions.PerforceWorkspaceNotFoundError`: If the workspace was not found
        :return: Perforce workspace specification template dictionary
        :rtype: Optional[dict]
        """

        port, user, workspace = cls.get_perforce_settings_from_ue_source_control()
        if not all([port, user, workspace]):
            raise exceptions.SourceControlProviderError(
                f'Failed to fetch Source control settings. Port: {port}, User: {user}, Workspace: {workspace}'
            )

        workspace_specification = perforce.get_perforce_workspace_specification(port=port, user=user, client=workspace)
        if not workspace_specification:
            raise exceptions.PerforceWorkspaceNotFoundError(
                'No Perforce workspace found. Please check P4 environment and try again'
            )

        workspace_name = workspace_specification['Client']
        workspace_root = workspace_specification['Root']

        workspace_name_template = 'workspace_name'.join('{}')

        workspace_specification_template = {
            'Client': workspace_name_template,
            'Root': workspace_root
        }

        if workspace_specification.get('Stream'):
            workspace_specification_template['Stream'] = workspace_specification['Stream']
        elif workspace_specification.get('View'):
            view_regex = rf'.*(\/\/{workspace_name}\/).*'
            view_templates = []
            for view in workspace_specification['View']:
                match = re.match(view_regex, view)
                if match and len(match.groups()) == 1 and match.groups()[0] == f'//{workspace_name}/':
                    view_templates.append(
                        view.replace(f'//{workspace_name}/', f'//{workspace_name_template}/')
                    )
                else:
                    view_templates.append(view)
            workspace_specification_template['View'] = view_templates

        return workspace_specification_template

    def _save_perforce_specification_template_file(self) -> Optional[str]:
        """
        Save the perforce specification template to the JSON file and return its path if exists

        :return: the path of the perforce specification template JSON file
        :rtype: Optional[str]
        """

        if not self._perforce_workspace_specification_template:
            return

        save_path = f'{unreal_utils.saved_dir}/perforce_specification_template.json'
        with open(save_path, 'w') as f:
            json.dump(self._perforce_workspace_specification_template, f, indent=4)

        return save_path

    def _resolve_task_template(self) -> str:
        """
        Build task (cmd line) from job template property of unreal.ConductorMoviePipelineExecutorJob
        and build context :class:`ciounreal.common.context.Context`

        Override parent task template (:meth:`ciounreal.conductor_job.conductor_job.ConductorJob._resolve_task_template()`)
        and add two commands:

        - Create P4 workspace before launch UE command
        - Delete P4 workspace after launch UE command

        Example cmd line:

        ``& %CIO_DIR%/p4utilsforunreal/cli.py create_workspace -UnrealProjectRelativePath={ue_project_relative_path}
        -PerforceWorkspaceSpecificationTemplate={_perforce_workspace_specification_template_path}
        -PerforceServer={perforce_server};
        UnrealEditor-Cmd.exe "{project_path}" -windowed -log -stdout -FullStdOutLogOutput -unattended
        -RenderOffscreen -noloadingscreen -allowstdoutlogverbosity -QueueManifest="{queue_manifest}"
        -LevelSequence="{level_sequence}" -Level="{map_path}" -execcmds="r.HLOD 0, py mrq_cli.py";
        & %CIO_DIR%/p4utilsforunreal/cli.py delete_workspace -UnrealProjectRelativePath={ue_project_relative_path}``

        :return: Job task as string
        :rtype: str
        """

        unreal_task_template = super()._resolve_task_template()

        unreal_project_path = unreal_utils.get_project_file_path().replace('\\', '/')
        workspace_root = self._perforce_workspace_specification_template['Root'].replace('\\', '/')
        unreal_project_relative_path = unreal_project_path.replace(workspace_root + '/', '')

        perforce_server = (self._mrq_job.conductor_settings.perforce_settings.perforce_server
                           or perforce.PerforceConnection().p4.port)

        unreal_python_path = self.get_unreal_python_path()

        p4utils_for_unreal_cli_path = f'{os.getenv("CIO_DIR")}/p4utilsforunreal/cli.py'.replace('\\', '/')

        # p4utils_exe = f'{unreal_utils.get_project_directory()}' \
        #               f'/Plugins/Conductor/Content/Python/p4utils-for-unreal.exe'.replace('\\', '/')

        create_workspace_template = \
            f'& "{unreal_python_path}" "{p4utils_for_unreal_cli_path}" create_workspace ' \
            f'-UnrealProjectRelativePath="{unreal_project_relative_path}" ' \
            f'-PerforceWorkspaceSpecificationTemplate="{self._perforce_workspace_specification_template_path}" ' \
            f'-PerforceServer="{perforce_server}"'

        delete_workspace_template = \
            f'& "{unreal_python_path}" "{p4utils_for_unreal_cli_path}" delete_workspace ' \
            f'-UnrealProjectRelativePath="{unreal_project_relative_path}"'

        # Save last exit code after Unreal process is complete
        save_ue_exit_code_template = f'$UE_EXIT_CODE = $LASTEXITCODE'

        # After workspace is deleted, exit with UE code if its bad (!=0), exit with last operation code otherwise
        exit_template = 'if ($UE_EXIT_CODE -ne 0) {exit $UE_EXIT_CODE} else {exit $LASTEXITCODE}'

        task_template = f'{create_workspace_template}; ' \
                        f'{unreal_task_template}; ' \
                        f'{save_ue_exit_code_template}; ' \
                        f'{delete_workspace_template}; ' \
                        f'{exit_template}'

        return task_template

    def _get_environment(self) -> PackageEnvironment:
        """"
        Build and return the :class:`ciocore.package_environment.PackageEnvironment` object that contains:

        - Environment variables from current software package
          (see :meth:`ciounreal.conductor_job.conductor_job.ConductorJob.get_current_software_package()`)
        - Extra environments from unreal.ConductorMoviePipelineExecutorJob
        - CIO_DIR environment variable if existed (point to the directory, where ciounreal package lives)
        - P4PORT, P4USER, P4PASSWD environment variables

        :return: :class:`ciocore.package_environment.PackageEnvironment` instance
        :rtype: :class:`ciocore.package_environment.PackageEnvironment`
        """
        environment = super()._get_environment()

        p4_vars = {"P4PORT": self.perforce_settings.perforce_server,
                    "P4USER": self.perforce_settings.perforce_username,
                    "P4PASSWD": self.perforce_settings.perforce_password}
            
        for p4_var, p4_value in p4_vars.items():
            try:
                env_list = [{'name': p4_var,
                            'value': p4_value,
                            'merge_policy': 'exclusive'
                            }]
                environment.extend(env_list)
            except ValueError:
                raise Exception(f"Invalid variable, please remove '{p4_var}' from your Conductor Environment Settings.")

        return environment

    def _get_upload_paths(self) -> list[str]:
        """
        Build and return the paths list to upload to the CoreWeave cloud storage. Include next path types:

        - Path described in `CIO_DIR` environment variable (Optional, if variable and path existed)
        - Path described in `UE_CONDUCTOR_PLUGIN_PATH` environment variable (Optional, if variable and path existed)
        - MoviePipelineQueue manifest file (see :meth:`ciounreal.conductor_job.conductor_job.ConductorJob._save_queue_manifest_file()`
        - Extra uploads (see :meth:`ciounreal.conductor_job.conductor_job.ConductorJob._get_extra_uploads()`
        - P4 Workspace specification template file path (see :meth:`ciounreal.conductor_job.perforce_conductor_job.PerforceConductorJob.get_perforce_workspace_specification_template()`)

        :return: List of local OS upload paths
        :rtype: list[str]
        """
        upload_paths = []

        if self._queue_manifest_file_path:
            upload_paths.append(self._queue_manifest_file_path)

        upload_paths.extend(self._get_extra_uploads())

        if os.getenv('CIO_DIR') and os.path.exists(os.getenv('CIO_DIR')):
            upload_paths.append(os.environ['CIO_DIR'])

        if os.getenv('UE_CONDUCTOR_PLUGIN_PATH') and os.path.exists(os.getenv('UE_CONDUCTOR_PLUGIN_PATH')):
            upload_paths.append(os.environ['UE_CONDUCTOR_PLUGIN_PATH'])

        upload_paths.append(self._perforce_workspace_specification_template_path)

        job_dependencies = self._get_mqr_job_dependency_paths(sync_missing=False)
        job_dependencies_descriptor_path = f'{unreal_utils.get_project_directory()}/Saved/{self._mrq_job.job_name}_dependencies.json'
        with open(job_dependencies_descriptor_path, 'w') as f:
            json.dump({'job_dependencies': job_dependencies}, f, indent=4)

        upload_paths.append(job_dependencies_descriptor_path)

        self.environment.extend(
            env_list=[
                {
                    'name': 'JOB_DEPENDENCIES_DESCRIPTOR_PATH',
                    'value': job_dependencies_descriptor_path,
                    'merge_policy': 'exclusive'
                }
            ]
        )

        return upload_paths



import sys
import logging


logger = None


def get_logger(reset=False, stream=sys.stdout, formatted=False) -> logging.Logger:
    """
    Returns logger for p4utils-for-unreal operations. Creates
    a new logger if one does not already exist for the app.

    :param reset: Remove all stream handlers and replaces them with a new one
    :param stream: Logger stream receiving handler logs
    :return logging.Logger:
    """
    global logger
    if logger is None:
        logger = logging.getLogger('p4utils-for-unreal')
        logger.setLevel(logging.DEBUG)
        add_handler(stream, formatted)

    if reset:
        reset_stream(stream, formatted)

    return logger

def reset_stream(stream, formatted):
    """
    Resets logger stream. May be used to prevent inactive 
    sys.stdout stream issue when DCC closes.
    """
    global logger

    while logger.handlers:
        handler = logger.handlers[0]
        handler.close
        logger.removeHandler(handler)

    logger.setLevel(logging.DEBUG)
    add_handler(stream, formatted)

def get_formatter():
    return logging.Formatter(
                fmt='[{asctime}][{levelno}] {name}: {message}',
                datefmt='%m/%d/%Y %H:%M:%S',
                style='{'
            )

def add_handler(stream, formatted):
    global logger
    sh = logging.StreamHandler(stream)
    if formatted:
        sh.setFormatter(get_formatter())
    logger.addHandler(sh)


#  Copyright 2024 CoreWeave
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

from typing import Optional

from ciocore import data as coredata
from ciocore.hardware_set import HardwareSet
from ciocore.package_tree import PackageTree

coredata.init('unrealengine', platforms=["windows"])


class ConductorData:
    """
    Helper class for getting useful data from the Conductor API
    """

    @staticmethod
    def get_ciocore_data() -> dict:
        """
        Get all the data from the Conductor API. Filter to only 
        receive singleton GPU instance types. 

        :return: projects, instance types and software package data
        :rtype: dict
        """
        return coredata.data(force=True, instances_filter="gpu.gpu_count=eq:1:int")

    @staticmethod
    def get_projects() -> list[str]:
        """
        Get projects from the Conductor API

        :return: list of the project names
        :rtype: list[str]
        """

        return coredata.data()['projects']

    @staticmethod
    def get_software() -> PackageTree:
        """
        Get software package data from the Conductor API

        :return: software package data object
        :rtype: :class:`ciocore.package_tree.PackageTree`
        """

        return coredata.data().get('software')

    @staticmethod
    def get_supported_software_packages() -> list[dict]:
        """
        Get software package data from the Conductor API
        that supported current product with which conductor was initialized

        :return: list of deserialized software package data objects
        :rtype: list[dict]
        """

        packages = []
        package_tree: PackageTree = ConductorData.get_software()
        if package_tree:
            for host_name in package_tree.supported_host_names():
                package: dict = package_tree.find_by_name(host_name)
                if package:
                    packages.append(package)
        return packages

    @staticmethod
    def get_instance_types() -> list[dict]:
        """
        Get list of instance types that are suitable for running UE render: GPU instances on Windows

        :return: list of instance type descriptors
        :rtype: list[dict]
        """

        result = []
        hardware: HardwareSet = coredata.data().get('instance_types')
        for category in hardware.categories:
            for entry in category['content']:
                if not (entry['operating_system'] == 'windows' and entry.get('gpu')):
                    continue
                result.append(entry)
        return result

    @staticmethod
    def get_instance_type_by_description(instance_type_description: str) -> Optional[dict]:
        """
        Get instance type by the given description

        :param instance_type_description: instance type description, for example 64 core 240GB Mem
               (4 T4 Tensor GPUs 16GB Mem)
        :return: instance type descriptor or None
        :rtype: Optional[dict]
        """

        return next(
            (
                instance_type for instance_type
                in ConductorData.get_instance_types()
                if instance_type['description'] == instance_type_description
            ),
            None
        )



#  Copyright 2024 CoreWeave
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import os
import sys
import uuid
import json
import pprint
import socket
from pathlib import Path
from typing import Optional

from p4utilsforunreal import perforce
from p4utilsforunreal.log import get_logger


logger = get_logger()


def get_conductor_task_id() -> Optional[str]:
    """
    Trying to find Conductor Task ID in os environment variable CONDUCTOR_TASK

    :return: Conductor Task ID or None
    :rtype: Optional[str]
    """

    return os.getenv('CONDUCTOR_TASK')


def get_workspace_name(project_name: str) -> str:
    """
    Build and return the workspace name based on the given project name:
    <USERNAME>_<HOST>_<PROJECT_NAME>_conductor

    :param project_name: Name of the project

    :return: Workspace name
    :rtype: str
    """

    unique_id = get_conductor_task_id() or 'NoConductorTaskId'
    return f'{os.getlogin()}_{socket.gethostname()}_{unique_id}_{project_name}_conductor'


def get_workspace_specification_template_from_file(workspace_specification_template_path: str) -> dict:
    """
    Read the given workspace specification template file path and return loaded content

    :param workspace_specification_template_path: Path to the workspace specification template file

    :return: Loaded workspace specification template dictionary
    :rtype: dict
    """

    if not os.path.exists(workspace_specification_template_path):
        raise FileNotFoundError(
            f'The workspace specification template does not exist: {workspace_specification_template_path}'
        )

    logger.info(f'Getting workspace specification template from file: {workspace_specification_template_path} ...')
    with open(workspace_specification_template_path, 'r') as f:
        return json.load(f)


def create_perforce_workspace_from_template(
        specification_template: dict,
        project_name: str,
        perforce_server: str,
        overridden_workspace_root: str = None,
) -> perforce.PerforceClient:
    """
    Creates Perforce workspace from the template

    :param specification_template: Workspace specification template dictionary
    :param project_name: Name of the project to build workspace name
    :param perforce_server: Perforce server address to connect to
    :param overridden_workspace_root: Workspace local path root (Optional, root from template is used by default)

    :return: :class:`p4utilsforunreal.perforce.PerforceClient` instance
    :rtype: :class:`p4utilsforunreal.perforce.PerforceClient`
    """

    logger.info(f'Creating perforce workspace from template: \n'
          f'Specification template: {specification_template}\n'
          f'Project: {project_name}\n'
          f'Perforce server: {perforce_server}\n'
          f'Overridden workspace root: {overridden_workspace_root}')

    workspace_name = get_workspace_name(project_name=project_name)

    specification_template_str = json.dumps(specification_template)
    specification_str = specification_template_str.replace('{workspace_name}', workspace_name)
    specification = json.loads(specification_str)

    if overridden_workspace_root:
        specification['Root'] = overridden_workspace_root

    logger.info(f'Specification: {specification}')

    perforce_client = perforce.PerforceClient(
        connection=perforce.PerforceConnection(port=perforce_server),
        name=workspace_name,
        specification=specification
    )

    perforce_client.save()

    logger.info('Perforce workspace created!')
    logger.info(pprint.pformat(perforce_client.spec))

    return perforce_client


def initial_workspace_sync(workspace: perforce.PerforceClient, unreal_project_relative_path: str) -> None:
    """
    Do initial workspace synchronization:

    - .uproject file
    - Binaries folder
    - Config folder
    - Plugins folder

    :param workspace: p4utilsforunreal.perforce.PerforceClient instance
    :param unreal_project_relative_path: path to the .uproject file relative to the workspace root
    """

    logger.info('Workspace initial synchronizing ...')

    workspace_root = workspace.spec['Root'].replace('\\', '/')

    paths_to_sync = [f'{workspace_root}/{unreal_project_relative_path}']

    unreal_project_directory = os.path.dirname(unreal_project_relative_path)

    for folder in ['Binaries', 'Config', 'Plugins']:
        tokens = filter(
            lambda t: t not in [None, ''],
            [workspace_root, unreal_project_directory, folder, '...']
        )
        paths_to_sync.append('/'.join(tokens))

    logger.info(f'Paths to sync: {paths_to_sync}')

    for path in paths_to_sync:
        try:
            workspace.sync(path)
        except Exception as e:
            logger.info(f'Initial workspace sync exception: {str(e)}')


def configure_project_source_control_settings(workspace: perforce.PerforceClient, unreal_project_relative_path: str):
    """
    Configure SourceControl settings (Saved/Config/WindowsEditor/SourceControlSettings.ini)
    with the current P4 connection settings

    :param workspace: p4utilsforunreal.perforce.PerforceClient instance
    :param unreal_project_relative_path: path to the .uproject file relative to the workspace root
    """

    logger.info('Configuring Unreal project SourceControl settings ...')
    unreal_project_directory = os.path.dirname(unreal_project_relative_path)
    tokens = filter(
        lambda t: t not in [None, ''],
        [workspace.spec['Root'], unreal_project_directory, 'Saved/Config/WindowsEditor/SourceControlSettings.ini']
    )
    source_control_settings_path = '/'.join(tokens)
    try:
        workspace.sync(source_control_settings_path)
    except:
        pass
    os.makedirs(os.path.dirname(source_control_settings_path), exist_ok=True)
    logger.info(f'Source Control settings file: {source_control_settings_path}')

    source_control_settings_lines = [
        '[PerforceSourceControl.PerforceSourceControlSettings]\n',
        'UseP4Config = False\n',
        f'Port = {workspace.p4.port}\n',
        f'UserName = {workspace.p4.user}\n',
        f'Workspace = {workspace.p4.client}\n\n',

        '[SourceControl.SourceControlSettings]\n',
        'Provider = Perforce\n'

    ]
    logger.info('source control settings:\n')
    for setting_line in source_control_settings_lines:
        logger.info(setting_line)

    with open(source_control_settings_path, 'w+') as f:
        for setting_line in source_control_settings_lines:
            f.write(setting_line)


def create_workspace(
        perforce_specification_template_path: str,
        unreal_project_relative_path: str,
        perforce_server: str,
        overridden_workspace_root: str = None
):
    """
    Create P4 workspace and execute next steps:

    - :meth:`p4utilsforunreal.app.get_workspace_specification_template_from_file()`
    - :meth:`p4utilsforunreal.app.initial_workspace_sync()`
    - :meth:`p4utilsforunreal.app.configure_project_source_control_settings()`

    :param perforce_specification_template_path: Path to the perforce specification template file to read specification from
    :param unreal_project_relative_path: path to the .uproject file relative to the workspace root
    :param perforce_server: Perforce server address to connect to
    :param overridden_workspace_root: Workspace local path root (Optional, root from template is used by default)
    """

    logger.info('Creating workspace with the following settings:\n'
                f'Specification template: {perforce_specification_template_path}\n'
                f'Unreal project relative path: {unreal_project_relative_path}'
                f'Perforce server: {perforce_server}'
                f'Overridden workspace root: {overridden_workspace_root}')

    workspace_specification_template = get_workspace_specification_template_from_file(
        workspace_specification_template_path=perforce_specification_template_path
    )

    workspace = create_perforce_workspace_from_template(
        specification_template=workspace_specification_template,
        project_name=Path(unreal_project_relative_path).stem,
        perforce_server=perforce_server,
        overridden_workspace_root=overridden_workspace_root
    )

    initial_workspace_sync(
        workspace=workspace,
        unreal_project_relative_path=unreal_project_relative_path
    )

    configure_project_source_control_settings(
        workspace=workspace,
        unreal_project_relative_path=unreal_project_relative_path
    )


def delete_workspace(unreal_project_relative_path: str):
    """
    Clear workspace files that are in depot and delete the workspace

    :param unreal_project_relative_path: path to the .uproject file relative to the workspace root
    """
    logger = get_logger(reset=True, stream=sys.stderr, formatted=True)
    
    project_name = Path(unreal_project_relative_path).stem

    logger.info(f'Deleting workspace for the project: {project_name}')

    workspace_name = get_workspace_name(project_name)
    p4 = perforce.PerforceConnection().p4

    last_exception = None

    workspace_root = p4.fetch_client(workspace_name).get('Root').replace('\\', '/')
    if workspace_root and os.path.exists(workspace_root):
        try:
            logger.info('Reverting changes in default changelist')
            p4.client = workspace_name
            p4.run('revert', '-c', 'default', workspace_root + '/...')
        except Exception as e:
            if 'file(s) not opened on this client' in str(e):
                logger.info('Nothing to revert')
                pass
            else:
                logger.info(f'Error handled while reverting changes: {e}')
                last_exception = e
        try:
            logger.info(f'Clearing workspace root: {workspace_root}')
            p4.client = workspace_name
            p4.run('sync', '-f', workspace_root + '/...#0')
        except Exception as e:
            if 'file(s) up-to-date' in str(e):
                logger.info('Nothing to clear')
            else:
                logger.info(f'Error handled while clearing workspace: {e}')
                last_exception = e

    try:
        logger.info(f'Deleting workspace: {workspace_name}')
        p4.run('client', '-d', workspace_name)
    except Exception as e:
        logger.info(f'Error handled while deleting workspace: {e}')
        last_exception = e

    if last_exception and isinstance(last_exception, Exception):
        raise last_exception



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import os
import json
import shlex
import shutil
import subprocess
import time

import setuptools
from setuptools.command.build_py import build_py

NICE_NAME = "Conductor for Unreal"
NAME = "ciounreal"
DESCRIPTION = "Unreal plugin for Conductor Cloud Rendering Platform."
URL = "https://github.com/ConductorTechnologies/ciounreal"
EMAIL = "info@conductortech.com"
AUTHOR = "Conductor"
REQUIRED = [
    "ciocore>=9.0.0,<10.0.0",
    "cioseq>=0.1.14,<1.0.0",
    "p4python==2024.1.2625398"
]
HERE = os.path.abspath(os.path.dirname(__file__))
DEV_VERSION = "dev.999"
UNREAL_BASE_PATH = os.getenv("UNREAL_BASE_PATH") or r"C:\Program Files\Epic Games"


with open(os.path.join(HERE, 'VERSION')) as version_file:
    VERSION = version_file.read().strip()


with open(os.path.join(HERE, 'README.md')) as readme:
    long_description = readme.read().strip()

long_description += "\n\n## Changelog\n\n"
with open(os.path.join(HERE, 'CHANGELOG.md')) as changelog:
    long_description += changelog.read().strip()


class UnrealPlugin:
    def __init__(self, version: str, runuat_path: str = None):
        self.version = version
        self.runuat_path = runuat_path

    def build(self, target: str):
        abs_target = os.path.abspath(target)
        source_folder = os.path.join(os.path.abspath(target), "Plugins", "Conductor")
        source_folder = source_folder.replace("\\", "/")

        target_folder = os.path.join(abs_target, "UAT", "Plugins", "Conductor")
        target_folder = target_folder.replace("\\", "/")

        build_command = f'"{self.runuat_path}" BuildPlugin ' \
                        f'-Plugin={source_folder}/Conductor.uplugin ' \
                        f'-Package={target_folder} ' \
                        f'-Rocket'
        process = subprocess.run(shlex.split(build_command))

        if process.returncode == 0:
            self.version_to_uplugin(f'{target_folder}/Conductor.uplugin')

            destination_zip = os.path.join(target, "Plugins", f"Conductor{self.version}")
            destination_dir = os.path.dirname(destination_zip)
            os.makedirs(destination_dir, exist_ok=True)
            shutil.make_archive(base_name=destination_zip, format='zip', root_dir=target_folder)
        else:
            print(f"Build failed with exitcode={process.returncode}. Check logs above for details")

        try:
            shutil.rmtree(os.path.join(target, "UAT"))
        except:
            pass

    @staticmethod
    def get_unreal_engine_plugins():
        if not os.path.exists(UNREAL_BASE_PATH):
            print("Search path for Unreal Engine not found. Skipping Unreal Engine plugin build.")
            return []

        engine_versions = [
            d for d in os.listdir(UNREAL_BASE_PATH)
            if os.path.isdir(os.path.join(UNREAL_BASE_PATH, d)) and d.startswith("UE_5")
        ]
        unreal_plugins = [
            UnrealPlugin(
                engine_version,
                os.path.join(UNREAL_BASE_PATH, engine_version, "Engine", "Build", "BatchFiles", "RunUAT.bat")
            )
            for engine_version in engine_versions
        ]
        return unreal_plugins

    @classmethod
    def version_to_uplugin(cls, uplugin_file: str):
        merge_data = {
            "IsBetaVersion": "b" in VERSION.lower(),
            "VersionName": VERSION,
            "FriendlyName" : NICE_NAME,
            "Version": 1,
            "Description": "Submit jobs to to the Conductor cloud rendering service.",
            "Category": "Rendering",
            "CreatedBy": AUTHOR,
            "CreatedByURL": "https://www.conductortech.com",
            "DocsURL": "https://docs.conductortech.com",
            "SupportURL": "https://support.conductortech.com/hc/en-us/requests/new",
            "EnabledByDefault": True
        }
        # uplugin_file = os.path.join(target_dir, "Plugins",  "Conductor", "Conductor.uplugin")
        with open(uplugin_file, "r") as f:
            uplugin = json.load(f)

        uplugin.update(merge_data)
        with open(uplugin_file, "w") as f:
            json.dump(uplugin, f, indent=4)

        print(f"Updated {uplugin_file} with version {VERSION}")
        print(f"Result uplugin: {uplugin}")


class BuildCommand(build_py):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def run(self):
        build_py.run(self)
        if not self.dry_run:
            target_dir = os.path.join(self.build_lib, NAME)
            self.write_version_info(target_dir)

            # self.version_to_uplugin(self.build_lib)

            unreal_plugins = UnrealPlugin.get_unreal_engine_plugins()
            for unreal_plugin in unreal_plugins:
                unreal_plugin.build(self.build_lib)

            self.remove_unreal_plugin_source(self.build_lib)

    @classmethod
    def remove_unreal_plugin_source(cls, target_dir):
        abs_target = os.path.abspath(target_dir)
        source_folder = os.path.join(abs_target, "Plugins", "Conductor")
        shutil.rmtree(source_folder)

    @classmethod
    def write_version_info(cls, target_dir: str):
        with open(os.path.join(target_dir, "VERSION"), "w") as f:
            f.write(VERSION)


def gather_package_files(directory):
    paths = []
    for (path, _, filenames) in os.walk(directory):
        for filename in filenames:
            paths.append(os.path.join("..", path, filename))
    return paths


extra_files = gather_package_files("Plugins")
extra_files.extend(gather_package_files("p4utilsforunreal"))
extra_files.extend(
    [
        '..\\VERSION',
        '..\\*.md'
    ]
)


setuptools.setup(
    author=AUTHOR,
    author_email=EMAIL,
    classifiers=[
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Topic :: Multimedia :: Graphics :: 3D Rendering",
    ],
    cmdclass={"build_py": BuildCommand},
    description=DESCRIPTION,
    install_requires=REQUIRED,
    long_description=long_description,
    long_description_content_type="text/markdown",
    name=NAME,
    package_dir={"": "."},
    packages=setuptools.find_packages(where="."),
    options={"bdist_wheel": {"universal": True}},
    include_package_data=True,
    url=URL,
    version=VERSION,
    zip_safe=False,
    package_data={NAME: extra_files}
)



#  Copyright 2024 CoreWeave
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
import json
import os
import re
import sys
import shlex
import subprocess
import errno
import shutil
import zipfile
from pathlib import Path


def fslash(path):
    return path.replace("\\", "/")


PACKAGE_DIR = fslash(os.path.dirname(os.path.abspath(__file__)))  # ciounreal
CIO_DIR = os.path.dirname(PACKAGE_DIR)  # ~/Conductor/unreal
HOME = fslash(os.path.expanduser("~"))
PLUGIN_NAME = "Conductor"

BASE_PATH = os.getenv("UNREAL_BASE_PATH") or r"C:\Program Files\Epic Games"
UNREAL_LAUNCHER_DATA = os.path.join(
    os.getenv("ProgramData", r"C:\ProgramData"), "Epic", "UnrealEngineLauncher", "LauncherInstalled.dat"
)
UNREAL_VERSIONS = ["5.2", "5.3", "5.4", "5.5", "5.6", "5.7"]

INIT_SCRIPT_CONTENT = f"""
import sys
import os

CIO_DIR = "{CIO_DIR}"
os.environ['CIO_DIR'] = CIO_DIR

if CIO_DIR not in sys.path:
    sys.path.append(CIO_DIR)

"""


def main(project_path=None):
    ue_plugin_directories = []
    if project_path:
        if not os.path.exists(project_path):
            print(f"The provided project path does not exist: {project_path}")
            print("Plugin will not be installed.")
            return
        project_plugins = UnrealPlugin.get_unreal_project_plugins(project_path)
        project_plugins.install_conductor()
    else:
        install_ciounreal()

        engine_plugins_list = UnrealPlugin.get_unreal_engine_plugins()
        for engine_plugins in engine_plugins_list:
            engine_plugins.install_conductor()

    print("Completed Unreal tool setup!")


def ensure_directory(directory):
    try:
        os.makedirs(directory)
    except OSError as ex:
        if ex.errno == errno.EEXIST and os.path.isdir(directory):
            pass
        else:
            raise

def install_ciounreal():
    #  Loop through each UE Version and see if directory exists
    for ue_ver in UNREAL_VERSIONS:
        ue_dir = "UE_{}".format(ue_ver)
        ue_dir_path = os.path.join(BASE_PATH, ue_dir)

        # If the directory exists, pip install ciounreal with UE native python
        if os.path.isdir(ue_dir_path):
            ue_python_path = os.path.join(ue_dir_path, "Engine", "Binaries", "ThirdParty", "Python3", "Win64", "python.exe")
            try:
                cmd = "\"{}\" -m pip install ciounreal --target \"{}\" --verbose".format(ue_python_path, CIO_DIR) 
                subprocess.run(shlex.split(cmd))
                print("Successful ciounreal install to {}!".format(ue_dir))
            except Exception:
                print("Failed ciounreal install to {}. Continuing...".format(ue_dir))
                continue

class UnrealPlugin:
    def __init__(self, version: str, plugins_path: str = None):
        self.version = version
        self.plugins_path = plugins_path

    def install_conductor(self):
        zipped = os.path.join(CIO_DIR, "Plugins", f"Conductor{self.version}.zip")
        zipped = zipped.replace("\\", "/")
        target = os.path.join(self.plugins_path, "Conductor")

        print(f"Installing Conductor Plugin {self.version} to {target}")
        # Create or empty a folder for the plugin.
        try:
            if os.path.isdir(target):
                shutil.rmtree(target)
        except Exception as e:
            print(f"Failed to remove existing {target}. Reason: {e}")
            return

        ensure_directory(target)
        with zipfile.ZipFile(zipped, "r") as zip_ref:
            zip_ref.extractall(target)

        with open(os.path.join(target, "Content", "Python", "init_unreal.py"), "r+") as f:
            content = f.read()
            f.seek(0, 0)
            content = INIT_SCRIPT_CONTENT + content
            f.write(content)

        print(f"Plugin installation completed! Conductor Plugin {self.version} installed to {target}")

    @staticmethod
    def get_base_path_engine_paths():
        if not os.path.exists(BASE_PATH):
            print("The Unreal Engine search base path does not exist.")
            return []
        engine_paths = [
            os.path.join(BASE_PATH, d) for d in os.listdir(BASE_PATH)
            if os.path.isdir(os.path.join(BASE_PATH, d)) and d.startswith("UE_5")
        ]
        return engine_paths

    @staticmethod
    def get_launcher_engine_paths():
        if not os.path.exists(UNREAL_LAUNCHER_DATA):
            print("Unreal Launcher data does not exist")
            return []

        with open(UNREAL_LAUNCHER_DATA, 'r') as f:
            ue_paths = []
            ue_launcher_installed = json.load(f)
            for app in ue_launcher_installed["InstallationList"]:
                if app["AppName"].startswith("UE_5") and app["NamespaceId"] == "ue":
                    ue_paths.append(app["InstallLocation"])
            return ue_paths

    @staticmethod
    def get_unreal_engine_plugins():
        launcher_engine_paths = [
            path.replace('/', '\\') for path in UnrealPlugin.get_launcher_engine_paths()
        ]
        base_engine_paths = [
            path.replace('/', '\\') for path in UnrealPlugin.get_base_path_engine_paths()
        ]
        all_engine_paths = list(set(launcher_engine_paths + base_engine_paths))

        unreal_plugins = [
            UnrealPlugin(
                os.path.basename(engine_path),
                os.path.join(engine_path, "Engine", "Plugins")
            )
            for engine_path in all_engine_paths
        ]
        print(f"Found {len(unreal_plugins)} Unreal Engine plugin directories")

        return unreal_plugins

    @staticmethod
    def get_unreal_project_plugins(project_path: str):
        uproject_path = next(
            (uproject_path for uproject_path in Path(project_path).glob('*.uproject')),
            None
        )
        if uproject_path:
            with open(uproject_path, "r") as f:
                uproject = json.load(f)
                engine_version = uproject.get("EngineAssociation")
                return UnrealPlugin("UE_" + engine_version, os.path.join(project_path, "Plugins"))
        return None


if __name__ == "__main__":
    # give a path to a project to install into. If not provided, it will install into the default Unreal Engine plugins directory.

    project_path = sys.argv[1] if len(sys.argv) > 1 else None
    main(project_path=project_path)



#  Copyright 2024 CoreWeave
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import os
import json
import socket
import unittest
from unittest.mock import patch, call, MagicMock

import pytest

from p4utilsforunreal.app import (
    get_workspace_specification_template_from_file,
    create_perforce_workspace_from_template,
    initial_workspace_sync,
    configure_project_source_control_settings
)
from p4utilsforunreal import perforce


def mocked_p4_sync(filepath: str = None):
    print(f'Mock sync file: {filepath}')


class TestP4UtilsForUnreal(unittest.TestCase):

    def test_get_perforce_specification_template_from_file(self):
        valid_content = dict(key1='value1', key2='value2')
        template_path = f'{os.path.dirname(__file__)}/specification_template.json'

        # Check valid template
        with open(template_path, 'w') as f:
            json.dump(valid_content, f)
        assert get_workspace_specification_template_from_file(template_path) == valid_content
        os.remove(template_path)

        # Check not existed template
        self.assertRaises(
            FileNotFoundError,
            get_workspace_specification_template_from_file,
            'not existed template.json'
        )

        # Check empty template
        open(template_path, 'w').close()
        self.assertRaises(
            json.decoder.JSONDecodeError,
            get_workspace_specification_template_from_file,
            template_path
        )
        os.remove(template_path)

    @patch('p4utilsforunreal.perforce.PerforceClient.save', new_callable=MagicMock)
    def test_create_perforce_workspace_from_template(self):
        spec = perforce.get_perforce_workspace_specification()
        spec_template = {
            'Client': '{workspace_name}',
            'Root': spec['Root'],
            'Stream': spec['Stream'],
            'View': [v.replace(spec['Client'], '{workspace_name}') for v in spec['View']]
        }

        project_name = 'CoreweaveConductorTest'

        client = create_perforce_workspace_from_template(
            specification_template=spec_template,
            project_name=project_name,
            perforce_server=perforce.PerforceConnection().p4.port
        )

        workspace_name = f'{os.getlogin()}_{socket.gethostname()}_{project_name}_conductor'

        assert client.spec['Client'] == workspace_name
        assert client.name == workspace_name
        assert client.spec['Stream'] == spec['Stream']
        assert client.spec['Root'] == spec['Root']
        for view in client.spec['View']:
            assert workspace_name in view

    @patch('p4utilsforunreal.perforce.PerforceClient.save', new_callable=MagicMock)
    @patch('p4utilsforunreal.perforce.PerforceClient.sync', side_effect=mocked_p4_sync)
    def test_initial_workspace_sync(self, mock_workspace_sync, mock_workspace_save):
        spec = perforce.get_perforce_workspace_specification()
        spec_template = {
            'Client': '{workspace_name}',
            'Root': spec['Root'],
            'Stream': spec['Stream'],
            'View': [v.replace(spec['Client'], '{workspace_name}') for v in spec['View']]
        }

        project_name = 'CoreweaveConductorTest'

        client = create_perforce_workspace_from_template(
            specification_template=spec_template,
            project_name=project_name,
            perforce_server=perforce.PerforceConnection().p4.port
        )

        initial_workspace_sync(client, 'ConductorTest/ConductorTest.uproject')
        mock_workspace_sync.assert_has_calls(
            [
                call(f'{client.spec["Root"]}/ConductorTest/ConductorTest.uproject'),
                call(f'{client.spec["Root"]}/ConductorTest/Binaries...'),
                call(f'{client.spec["Root"]}/ConductorTest/Config...'),
            ]
        )

        initial_workspace_sync(client, 'ConductorTest.uproject')
        mock_workspace_sync.assert_has_calls(
            [
                call(f'{client.spec["Root"]}/ConductorTest.uproject'),
                call(f'{client.spec["Root"]}/Binaries...'),
                call(f'{client.spec["Root"]}/Config...'),
            ]
        )



# Copyright 2024 CONDUCTOR TECHNOLOGIES. All Rights Reserved.

import os
import json
import socket
import unreal
import logging

from p4utilsforunreal import perforce, app

from ciounreal.common import unreal_utils
from ciounreal.common import dependency_collector
from ciounreal.render_executors import MoviePipelineConductorLocalEditorExecutor

# We need to keep the render executor here to make callbacks being executed after render is complete/failed
render_executor = None


def wait_for_asset_registry():
    unreal_utils.log('MRQ CLI: Waiting for asset registry completion ...')
    asset_registry = unreal.AssetRegistryHelpers.get_asset_registry()
    asset_registry.wait_for_completion()


def set_movie_pipeline_queue(movie_pipeline_queue: unreal.MoviePipelineQueue) -> unreal.MoviePipelineQueue:
    unreal_utils.log('MRQ CLI: setting Movie Pipeline Queue...')
    unreal_utils.log(f'MRQ CLI: original Movie Pipeline Queue: {movie_pipeline_queue}')

    # The queue subsystem behaves like a singleton so lear all the jobs in the current queue.
    pipeline_queue = unreal.get_editor_subsystem(unreal.MoviePipelineQueueSubsystem).get_queue()
    pipeline_queue.delete_all_jobs()

    # Replace the contents of this queue with a copy of the contents from another queue.
    pipeline_queue.copy_from(movie_pipeline_queue)

    unreal_utils.log(f'MRQ CLI: Resulted Movie Pipeline Queue: {pipeline_queue}')

    return pipeline_queue


def get_movie_pipeline_queue_from_manifest(mpq_manifest_path: str) -> unreal.MoviePipelineQueue:
    unreal_utils.log(f'MRQ CLI: Getting Movie Pipeline Queue from manifest: {mpq_manifest_path}')
    return unreal.MoviePipelineLibrary.load_manifest_file_from_string(mpq_manifest_path)


def executor_errored_callback(
        pipeline_executor: unreal.MoviePipelineExecutorBase,
        pipeline_with_error: unreal.MoviePipeline,
        is_fatal: bool,
        error_text: unreal.Text
):
    unreal_utils.log(f'''MoviePipelineConductorLocalEditorExecutor {pipeline_executor}: Error handled:
                     Movie Pipeline: {pipeline_with_error}
                     Fatal: {is_fatal}
                     Error: {error_text}''', logging.ERROR)

    global render_executor
    del render_executor

    unreal.SystemLibrary.quit_editor()


def executor_finished_callback(pipeline_executor: unreal.MoviePipelineExecutorBase, success: bool = None):
    unreal_utils.log(f'MoviePipelineConductorLocalEditorExecutor: Rendering is complete')
    if success is not None:
        unreal_utils.log(f'MoviePipelineConductorLocalEditorExecutor: Success: {success}')

    global render_executor
    del render_executor

    unreal.SystemLibrary.quit_editor()


def main():
    """
    UnrealEditor-Cmd.exe "{project_path}"
    -windowed -log -stdout -FullStdOutLogOutput -unattended -RenderOffscreen -noloadingscreen -allowstdoutlogverbosity
    -QueueManifest=”{queue_manifest}” -execcmds=”r.HLOD 0, py mrq_cli.py”
    """

    tokens, switchers, cmd_parameters = unreal.SystemLibrary.parse_command_line(
        unreal.SystemLibrary.get_command_line()
    )
    unreal_utils.log(f'''MRQ CLI: Parsed arguments:
                     Tokens: {tokens}
                     Switchers: {switchers}
                     CMD Parameters: {cmd_parameters}''')

    wait_for_asset_registry()

    original_queue = get_movie_pipeline_queue_from_manifest(cmd_parameters['QueueManifest'])
    pipeline_queue = set_movie_pipeline_queue(original_queue)

    global render_executor
    render_executor = MoviePipelineConductorLocalEditorExecutor()

    p4 = None

    for job in pipeline_queue.get_jobs():
        if job.conductor_settings.perforce_settings.use_perforce and not p4:
            p4_connect = perforce.PerforceConnection(required_connection=True)
            p4 = p4_connect.p4
            p4.client = app.get_workspace_name(project_name=unreal_utils.get_project_name())
            unreal_utils.log('MRQ CLI: Perforce connection established.')
        else:
            print (f"MRQ CLI: Skipping Perforce for the job: {job.job_name}")
    
        unreal_utils.log(f'MRQ CLI: Collecting dependencies for the job: {job.job_name}')
        deps_collector = dependency_collector.DependencyCollector()
        deps_collector.collect(
            asset_path=cmd_parameters['LevelSequence'],
            filter_method=dependency_collector.DependencyFilters.dependency_in_game_folder
        )
        deps_collector.collect(
            asset_path=cmd_parameters['Level'],
            filter_method=dependency_collector.DependencyFilters.dependency_in_game_folder
        )

        job_dependencies_descriptor_path = os.getenv('JOB_DEPENDENCIES_DESCRIPTOR_PATH', '')
        if os.path.exists(job_dependencies_descriptor_path):
            with open(job_dependencies_descriptor_path, 'r') as f:
                job_dependencies_data = json.load(f)
                for job_dependency in job_dependencies_data.get('job_dependencies', []):
                    if os.path.exists(job_dependency):
                        continue
                    synced = unreal.SourceControl.sync_files([job_dependency])
                    if synced:
                        unreal.AssetRegistryHelpers().get_asset_registry().scan_modified_asset_files([job_dependency])
                        unreal.AssetRegistryHelpers().get_asset_registry().scan_paths_synchronous(
                            [job_dependency], True, True
                        )

        if p4:
            job_extra_uploads_files = [p.file_path for p in job.conductor_settings.uploads_settings.files.paths]
            job_extra_uploads_folders = [p.path + '/...' for p in job.conductor_settings.uploads_settings.folders.paths]
            job_extra_uploads = job_extra_uploads_files + job_extra_uploads_folders

            for job_extra_upload in job_extra_uploads:
                try:
                    p4.run('sync', job_extra_upload)
                except Exception as e:
                    unreal_utils.log(f'Sync extra upload error: {str(e)}', logging.ERROR)
                    continue

        render_data_missed = False
        if not unreal.EditorAssetLibrary.does_asset_exist(cmd_parameters['LevelSequence']):
            render_data_missed = True
            unreal_utils.log(f'LevelSequence {cmd_parameters["LevelSequence"]} does not exist, cancel rendering')
        if not unreal.EditorAssetLibrary.does_asset_exist(cmd_parameters['Level']):
            render_data_missed = True
            unreal_utils.log(f'Level {cmd_parameters["Level"]} does not exist, cancel rendering')

        if render_data_missed:
            exit_code = 101
            unreal_utils.log(f'Level or/and LevelSequence assets are missing, exiting with exit code {exit_code}')
            unreal.ConductorMiscFunctionLibrary.request_exit(True, exit_code)

    # Add callbacks on complete and error actions to handle it and provide output
    render_executor.on_executor_finished_delegate.add_callable_unique(executor_finished_callback)
    render_executor.on_executor_errored_delegate.add_callable_unique(executor_errored_callback)

    unreal_utils.log('MRQ CLI: Starting executing Movie Pipeline Queue ...')
    render_executor.execute(pipeline_queue)


if __name__ == '__main__':
    main()



#  Copyright 2024 CoreWeave
#
#  Permission is hereby granted, free of charge, to any person obtaining a copy of this software and associated documentation files (the "Software"), to deal in the Software without restriction, including without limitation the rights to use, copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the Software is furnished to do so, subject to the following conditions:
#
#  The above copyright notice and this permission notice shall be included in all copies or substantial portions of the Software.
#
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

import os
import re
import unreal
from typing import Optional

from ciocore.package_environment import PackageEnvironment

from ciounreal import exceptions
from ciounreal.common import context
from ciounreal.common import unreal_utils
from ciounreal.common import conductor_data
from ciounreal.common import dependency_collector


class ConductorJob:
    """
    Conductor Job payload representation
    """

    def __init__(self, mrq_job: unreal.ConductorMoviePipelineExecutorJob = None, build_context: context.Context = None):
        if not mrq_job and not build_context:
            return

        self._mrq_job = mrq_job
        self._context = build_context

        self._queue_manifest_file_path = self._save_queue_manifest_file()

        self.project: str = self._mrq_job.conductor_settings.general_settings.project
        self.job_title: str = self._resolve_job_title()
        self.instance_type: dict = self._get_instance_type()
        self.pre_emptible: bool = self._mrq_job.conductor_settings.general_settings.pre_emptible
        self.software_package_ids: list[str] = [self.get_current_software_package()['package_id']]
        self.environment: PackageEnvironment = self._get_environment()
        self.output_path: str = self._resolve_output_path()
        self.location: str = self._mrq_job.conductor_settings.advanced_settings.location_tag
        self.local_upload: bool = False if self._mrq_job.conductor_settings.uploads_settings.use_daemon else True
        self.upload_paths: list[str] = self._get_upload_paths()
        self.tasks_data: list[dict] = [
            {'command': self._resolve_task_template()}
        ]
        self.notifications: list[str] = re.findall(
            pattern=r'[\w.-]+@[\w.-]+\.\w+',
            string=self._mrq_job.conductor_settings.advanced_settings.notify
        )

    @classmethod
    def from_dict(cls, data: dict) -> "ConductorJob":
        """
        Creates a :class:`ciounreal.conductor_job.conductor_job.ConductorJob` instance from the given data

        :param data: Dictionary containing information about the Conductor Job

        :return: :class:`ciounreal.conductor_job.conductor_job.ConductorJob` instance
        :rtype: :class:`ciounreal.conductor_job.conductor_job.ConductorJob`
        """

        conductor_job = cls(None, None)
        conductor_job.project = data['project']
        conductor_job.job_title = data['job_title']
        conductor_job.instance_type = data['instance_type']
        conductor_job.pre_emptible = data['preemptible']
        conductor_job.software_package_ids = data['software_package_ids']
        conductor_job.environment = data['environment']
        conductor_job.output_path = data['output_path']
        conductor_job.location = data['location']
        conductor_job.local_upload = data['local_upload']
        conductor_job.upload_paths = data['upload_paths']
        conductor_job.tasks_data = data['tasks_data']
        conductor_job.notifications = data['notify']

        return conductor_job

    @classmethod
    def get_current_software_package(cls) -> Optional[dict]:
        """
        Get the software package that fit the current UE version.

        Uses :meth:`ciounreal.common.conductor_data.ConductorData.get_supported_software_packages()`
        to get all supported software packages and :meth:`ciounreal.common.unreal_utils.get_engine_version_number()`
        to get the current UE version

        :raises: :class:`ciounreal.exceptions.SoftwareNotFoundError`: If no suitable software package was found
        :return: Software package descriptor or None
        :rtype: Optional[dict]
        """

        engine_version_number = unreal_utils.get_engine_version_number()
        major, minor, patch = engine_version_number.split('.')
        software_packages = conductor_data.ConductorData.get_supported_software_packages()
        current_software_package = next(
            (
                package for package in software_packages
                if package['major_version'] == major and package['minor_version'] == minor
            ),
            None
        )
        if not current_software_package:
            raise exceptions.SoftwareNotFoundError(
                f'No software found for the current version: {engine_version_number}'
            )
        return current_software_package

    def _get_instance_type(self) -> Optional[dict]:
        """
        Get the instance type by description provided in the unreal.ConductorMoviePipelineExecutorJob appropriate field

        :return: Instance type descriptor or None
        :rtype: Optional[dict]
        """

        return conductor_data.ConductorData.get_instance_type_by_description(
            instance_type_description=self._mrq_job.conductor_settings.general_settings.instance_type
        )

    def _get_environment(self) -> PackageEnvironment:
        """
        Build and return the :class:`ciocore.package_environment.PackageEnvironment` object that contains:

        - Environment variables from current software package
          (see :meth:`ciounreal.conductor_job.conductor_job.ConductorJob.get_current_software_package()`)
        - Extra environments from unreal.ConductorMoviePipelineExecutorJob
        - CIO_DIR environment variable if existed (point to the directory, where ciounreal package lives)

        :return: :class:`ciocore.package_environment.PackageEnvironment` instance
        :rtype: :class:`ciocore.package_environment.PackageEnvironment`
        """

        current_software_package = self.get_current_software_package()
        environment = PackageEnvironment(current_software_package)

        env_list: list[dict] = []
        extra_environment = self._mrq_job.conductor_settings.environment_settings.variables.key_values
        for key in extra_environment:
            env_list.append(
                {
                    'name': key,
                    'value': extra_environment[key].value,
                    'merge_policy': extra_environment[key].merge_policy or 'append'
                }
            )

        if os.getenv('CIO_DIR'):
            env_list.append(
                {
                    'name': 'CIO_DIR',
                    'value': os.environ['CIO_DIR'],
                    'merge_policy': 'exclusive'
                }
            )

        env_list.append(
            {
                    'name': 'CONDUCTOR_DISABLE_PLUGIN',
                    'value': '1',
                    'merge_policy': 'exclusive'
                }
        )

        environment.extend(env_list)

        return environment

    def _get_mrq_job_dependencies(self, sync_missing: bool = True) -> list[str]:
        """
        Build and return list of dependencies of level and level sequence from unreal.ConductorMoviePipelineExecutorJob

        :param sync_missing: Whether to sync missing dependencies or not

        :return: list of dependencies of level and level sequence
        :rtype: list[str]
        """

        deps_collector = dependency_collector.DependencyCollector()

        level_sequence_path = self._context['level_sequence']
        unreal_utils.log(f'Get Level Sequence dependencies: {level_sequence_path}')
        level_sequence_dependencies = deps_collector.collect(
            asset_path=level_sequence_path,
            filter_method=dependency_collector.DependencyFilters.dependency_in_game_folder,
            sync_missing=sync_missing
        )
        level_sequence_dependencies.append(level_sequence_path)

        level_path = self._context['map_path']
        unreal_utils.log(f'Get Level dependencies: {level_path}')
        level_dependencies = deps_collector.collect(
            asset_path=level_path,
            filter_method=dependency_collector.DependencyFilters.dependency_in_game_folder,
            sync_missing=sync_missing
        )
        level_dependencies.append(level_path)

        return list(set(level_sequence_dependencies + level_dependencies))

    def _get_extra_uploads(self) -> list[str]:
        """
        Build and return list of the extra uploads paths (files and folders)
        if any added in unreal.ConductorMoviePipelineExecutorJob

        :return: list of the extra uploads paths
        :rtype: list[str]
        """

        extra_upload_paths = []

        job_extra_uploads_files = [p.file_path for p in self._mrq_job.conductor_settings.uploads_settings.files.paths]
        job_extra_uploads_folders = [p.path for p in self._mrq_job.conductor_settings.uploads_settings.folders.paths]
        job_extra_uploads = job_extra_uploads_files + job_extra_uploads_folders

        for extra_upload in job_extra_uploads:
            if os.path.exists(extra_upload):
                extra_upload_paths.append(extra_upload)
            else:
                os_path = f'{unreal_utils.get_project_directory()}/{extra_upload}'.replace('\\', '/')
                if os.path.exists(os_path):
                    extra_upload_paths.append(os_path)

        return extra_upload_paths

    def _get_mqr_job_dependency_paths(self, sync_missing: bool = True) -> list[str]:
        """
        Get MRQ Job dependencies
        (see :meth:`ciounreal.conductor_job.conductor_job.ConductorJob._get_mrq_job_dependencies()`),
        convert dependencies UE paths (/Game/...) to the OS paths, filter not existed and return resulted list of paths

        :param sync_missing: Whether to sync missing dependencies or not

        :return: List of MRQ Job dependencies as local OS paths
        :rtype: list[str]
        """

        os_dependencies = []

        job_dependencies = self._get_mrq_job_dependencies(sync_missing)
        for dependency in job_dependencies:
            os_dependency = unreal_utils.os_path_from_unreal_path(dependency, with_ext=True)
            if os.path.exists(os_dependency):
                os_dependencies.append(os_dependency)

        return os_dependencies

    def _get_upload_paths(self) -> list[str]:
        """
        Build and return the paths list to upload to the CoreWeave cloud storage. Include next path types:

        - Binaries, Config and Plugins folders inside the project directory
        - Path described in `CIO_DIR` environment variable (Optional, if variable and path existed)
        - Path described in `UE_CONDUCTOR_PLUGIN_PATH` environment variable (Optional, if variable and path existed)
        - MoviePipelineQueue manifest file (see :meth:`ciounreal.conductor_job.conductor_job.ConductorJob._save_queue_manifest_file()`
        - Extra uploads (see :meth:`ciounreal.conductor_job.conductor_job.ConductorJob._get_extra_uploads()`
        - MRQ Job dependencies (see :meth:`ciounreal.conductor_job.conductor_job.ConductorJob._get_mqr_job_dependency_paths()`)

        :return: List of local OS upload paths
        :rtype: list[str]
        """

        upload_paths = [
            unreal_utils.get_project_file_path(),
            f'{unreal_utils.get_project_directory()}/Binaries',
            f'{unreal_utils.get_project_directory()}/Config',
            f'{unreal_utils.get_project_directory()}/Plugins'
        ]

        upload_paths = [p for p in upload_paths if os.path.exists(p)]

        if os.getenv('CIO_DIR') and os.path.exists(os.getenv('CIO_DIR')):
            upload_paths.append(os.environ['CIO_DIR'])

        if os.getenv('UE_CONDUCTOR_PLUGIN_PATH') and os.path.exists(os.getenv('UE_CONDUCTOR_PLUGIN_PATH')):
            upload_paths.append(os.environ['UE_CONDUCTOR_PLUGIN_PATH'])

        if self._queue_manifest_file_path:
            upload_paths.append(self._queue_manifest_file_path)

        upload_paths.extend(self._get_extra_uploads())

        upload_paths.extend(self._get_mqr_job_dependency_paths(sync_missing=False))

        return upload_paths

    def _resolve_job_title(self) -> str:
        """
        Build job title from job template property of unreal.ConductorMoviePipelineExecutorJob
        and build context :class:`ciounreal.common.context.Context`

        :return: Job title as string
        :rtype: str
        """

        return self._mrq_job.conductor_settings.general_settings.job_title.format_map(self._context)

    def _resolve_output_path(self) -> str:
        """
        Build output path from job output_directory property of unreal.ConductorMoviePipelineExecutorJob's configuration
        and build context :class:`ciounreal.common.context.Context`

        :return: Job output path as string
        :rtype: str
        """

        output_settings = self._mrq_job.get_configuration().find_setting_by_class(unreal.MoviePipelineOutputSetting)
        output_path: str = output_settings.output_directory.path.rstrip('/')
        return output_path.format_map(self._context)

    def _resolve_task_template(self) -> str:
        """
        Build task (cmd line) from job template property of unreal.ConductorMoviePipelineExecutorJob
        and build context :class:`ciounreal.common.context.Context`

        Example cmd line:

        ``UnrealEditor-Cmd.exe "{project_path}" -windowed -log -stdout -FullStdOutLogOutput -unattended
        -RenderOffscreen -noloadingscreen -allowstdoutlogverbosity -QueueManifest="{queue_manifest}"
        -LevelSequence="{level_sequence}" -Level="{map_path}" -execcmds="r.HLOD 0, py mrq_cli.py"``

        :return: Job task as string
        :rtype: str
        """

        task_template: str = self._mrq_job.conductor_settings.advanced_settings.template

        ctx = self._context.copy()
        ctx['queue_manifest'] = self._queue_manifest_file_path

        return task_template.format_map(ctx)

    def _save_queue_manifest_file(self) -> str:
        """
        Duplicate unreal.ConductorMoviePipelineExecutorJob object to the new unreal.MoviePipelineQueue object
        and save the queue as file

        :return: Path to the unreal.MoviePipelineQueue manifest file
        :rtype: str
        """

        new_queue = unreal.MoviePipelineQueue()
        new_queue.duplicate_job(self._mrq_job)

        # In duplicated job remove empty auto-detected files since we don't want them to be saved in manifest
        # List of the files is moved to OpenJob attachments
        # new_job.preset_overrides.job_attachments.input_files.auto_detected = unreal.ConductorFileAttachmentsArray()

        duplicated_queue, manifest_path = unreal.MoviePipelineEditorLibrary.save_queue_to_manifest_file(
            new_queue
        )
        manifest_path = unreal.Paths.convert_relative_path_to_full(manifest_path)
        return manifest_path

    def as_dict(self) -> dict:
        """
        Return instance as dictionary

        :return: :class:`ciounreal.conductor_job.conductor_job.ConductorJob` as dictionary
        :rtype: dict
        """

        return dict(
            project=self.project,
            job_title=self.job_title,
            instance_type=self.instance_type['name'],
            preemptible=self.pre_emptible,
            software_package_ids=self.software_package_ids,
            environment=dict(self.environment),
            output_path=self.output_path,
            location=self.location,
            local_upload=self.local_upload,
            upload_paths=self.upload_paths,
            tasks_data=self.tasks_data,
            notify=self.notifications
        )
